# -*- coding: utf-8 -*-

##############################################################################
#
# This file is part of Sardana
#
# http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
#
# Copyright: 2013 PetraIII,
#            2013 Synchrotron-Soleil
#                 L'Orme des Merisiers Saint-Aubin
#                 BP 48 91192 GIF-sur-YVETTE CEDEX
#
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

"""This module contains the definition of an Hkl pseudo motor controller
for the Sardana Device Pool"""

__author__ = ("Maria-Teresa Nunez-Pardo-de-Verra <tnunez@mail.desy.de>, "
              "Picca Frédéric-Emmanuel <picca@synchrotron-soleil.fr>")
__copyright__ = ("Copyright (c) 2013 DESY "
                 "Copyright (c) 2013 Synchrotron-Soleil "
                 "L'Orme des Merisiers Saint-Aubin "
                 "BP 48 91192 GIF-sur-YVETTE CEDEX")
__license__ = 'LGPL-3+'

__all__ = ["Diffrac6C", "DiffracE6C", "DiffracE4C", "Diffractometer"]

__docformat__ = 'restructuredtext'

import os
import time

import PyTango

from itertools import chain

from gi.repository import GLib
from gi.repository import Hkl

from taurus.core.util.codecs import CodecFactory

from sardana import DataAccess
from sardana.pool.controller import PseudoMotorController
from sardana.pool.controller import Type, Access, Description
from sardana.pool.controller import Memorize, Memorized, MemorizedNoInit, NotMemorized  # noqa

ReadOnly = DataAccess.ReadOnly
ReadWrite = DataAccess.ReadWrite
USER = Hkl.UnitEnum.USER
DEFAULT_CRYSTAL = "default_crystal"


from taurus.core.util.log import Logger

logger = Logger.getLogger("ControllerManager")

logger.info("The diffractometer controller is at early stage. Controller attributes and commands can slightly change.")


class AxisPar(object):

    def __init__(self, engine, name):
        self.engine = engine
        self.name = name
        self._modes = None

    @property
    def mode(self):
        return self.engine.current_mode_get()

    @mode.setter
    def mode(self, value):
        self.engine.current_mode_set(value)

    @property
    def modes(self):
        if self._modes is None:
            self._modes = self.engine.modes_names_get()
        return self._modes

    @property
    def modeparameters(self):
        return self.engine.parameters_names_get()

    @property
    def modeparametersvalues(self):
        return [p.value_get(USER)
                for p in [self.engine.parameter_get(n)
                          for n in self.modeparameters]]

    @modeparametersvalues.setter
    def modeparametersvalues(self, value):
        for parameter, v in zip(self.modeparameters,
                                value):
            p = self.engine.parameter_get(parameter)
            p.value_set(v, USER)
            self.engine.parameter_set(parameter, p)

# TODO: all the fit attributes (AFit or GammaFit)should change its type to bool


class DiffracBasis(PseudoMotorController):

    """ The PseudoMotor controller for the diffractometer"""

    ctrl_properties = {
        'DiffractometerType': {
            Type: str,
            Description: 'Type of the diffractometer, e.g. E6C'},
    }

    ctrl_attributes = {'Crystal': {Type: str,
                                   Memorize: MemorizedNoInit,  # noqa If Crystal is changed to memorized. The changes commented with "memcrystal"
                                   Access: ReadWrite},
                       'AffineCrystal': {Type: int,
                                         Memorize: MemorizedNoInit,
                                         Description: "Fine tunning of lattice parameters and UB matrix based on current crystal reflections. Reflections with affinement set to 0 are not used. A new crystal with the post fix (affine) is created and set as current crystal",  # noqa
                                         Access: ReadWrite},
                       'Wavelength': {Type: float,
                                      Memorize: MemorizedNoInit,
                                      Access: ReadWrite},
                       'EngineMode': {Type: str,  # TODO delete
                                      Memorize: MemorizedNoInit,
                                      Access: ReadWrite},
                       'EngineModeList': {Type: (str,), Access: ReadOnly},
                       # TODO delete
                       'HKLModeList': {Type: (str,), Access: ReadOnly},
                       'UBMatrix': {Type: ((float,), ),
                                    Description: "The reflection matrix",
                                    Access: ReadOnly},
                       'Ux': {Type: float, Access: ReadWrite},
                       'Uy': {Type: float, Access: ReadWrite},
                       'Uz': {Type: float, Access: ReadWrite},
                       'ComputeUB': {Type: (int,),
                                    Description: "Compute reflection matrix using two given reflections",  # noqa
                                    Access: ReadWrite},
                       'LatticeReciprocal': {Type: (float,),
                                             Description: "The reciprocal lattice parameters of the sample",  # noqa
                                             Access: ReadOnly},
                       'A': {Type: float,
                             Description: "a parameter of the lattice",
                             Memorize: MemorizedNoInit,
                             Access: ReadWrite},
                       'B': {Type: float,
                             Description: "b parameter of the lattice",
                             Memorize: MemorizedNoInit,
                             Access: ReadWrite},
                       'C': {Type: float,
                             Description: "c parameter of the lattice",
                             Memorize: MemorizedNoInit,
                             Access: ReadWrite},
                       'Alpha': {Type: float,
                                 Description: "alpha parameter of the lattice",
                                 Memorize: MemorizedNoInit,
                                 Access: ReadWrite},
                       'Beta': {Type: float,
                                Description: "beta parameter of the lattice",
                                Memorize: MemorizedNoInit,
                                Access: ReadWrite},
                       'Gamma': {Type: float,
                                 Description: "gamma parameter of the lattice",
                                 Memorize: MemorizedNoInit,
                                 Access: ReadWrite},
                       'AFit': {Type: int,
                                Description: "Fit value of the a parameter of the lattice",  # noqa
                                Memorize: MemorizedNoInit,
                                Access: ReadWrite},
                       'BFit': {Type: int,
                                Description: "Fit value of the b parameter of the lattice",  # noqa
                                Memorize: MemorizedNoInit,
                                Access: ReadWrite},
                       'CFit': {Type: int,
                                Description: "Fit value of the c parameter of the lattice",  # noqa
                                Memorize: MemorizedNoInit,
                                Access: ReadWrite},
                       'AlphaFit': {Type: int,
                                    Description: "Fit value of the alpha parameter of the lattice",  # noqa
                                    Memorize: MemorizedNoInit,
                                    Access: ReadWrite},
                       'BetaFit': {Type: int,
                                   Description: "Fit value of the beta parameter of the lattice",  # noqa
                                   Memorize: MemorizedNoInit,
                                   Access: ReadWrite},
                       'GammaFit': {Type: int,
                                    Description: "Fit value of the gamma parameter of the lattice",  # noqa
                                    Memorize: MemorizedNoInit,
                                    Access: ReadWrite},
                       'TrajectoryList': {Type: ((float,), (float,)),
                                          Description: "List of trajectories for hklSim",  # noqa
                                          Access: ReadOnly},
                       'SelectedTrajectory': {Type: int,
                                              Description: "Index of the trajectory you want to take for the given hkl values. 0 (by default) if first.",  # noqa
                                              Access: ReadWrite},
                       'ComputeTrajectoriesSim': {Type: (float,),
                                                  Description: "Pseudo motor values to compute the list of trajectories (1, 2 or 3 args)",  # noqa
                                                  Access: ReadWrite},
                       'Engine': {Type: str,
                                  Memorize: MemorizedNoInit,
                                  Access: ReadWrite},
                       'EngineList': {Type: (str,), Access: ReadOnly},
                       'EnginesConf': {Type: str, Access: ReadOnly},
                       'CrystalList': {Type: (str,), Access: ReadOnly},
                       'AddCrystal': {Type: str,
                                      Memorize: MemorizedNoInit,
                                      Access: ReadWrite},
                       'DeleteCrystal': {Type: str,
                                         Memorize: MemorizedNoInit,
                                         Access: ReadWrite},
                       'AddReflection': {Type: (float,),
                                         Description: "Add reflection to current sample",  # noqa
                                         Access: ReadWrite},
                       'SubstituteReflection': {Type: (float,),
                                                  Description: "Substitute reflexion with the given index by the given one",  # noqa
                                                  Access: ReadWrite},
                       'SwapReflections01': {Type: int,
                                             Description: "Swap primary and secondary reflections",  # noqa
                                             Access: ReadWrite},
                       'ReflectionList': {Type: ((float,), (float,)),
                                          Description: "List of reflections for current sample",  # noqa
                                          Access: ReadOnly},
                       'RemoveReflection': {Type: int,
                                            Memorize: MemorizedNoInit,
                                            Description: "Remove reflection with given index",  # noqa
                                            Access: ReadWrite},
                       'LoadReflections': {Type: str,
                                           Description: "Load the reflections from the file given as argument",  # noqa
                                           Memorize: MemorizedNoInit,
                                           Access: ReadWrite},
                       'SaveReflections': {Type: str,
                                           Description: "Save current reflections to file",  # noqa
                                           Memorize: MemorizedNoInit,
                                           Access: ReadWrite},
                       'LoadCrystal':     {Type: str,
                                           Description: "Load the lattice parameters and reflections from the file corresponding to the given crystal",  # noqa
                                           Memorize: MemorizedNoInit,
                                           Access: ReadWrite},
                       'SaveCrystal': {Type: int,
                                       Description: "Save current crystal parameters and reflections",  # noqa
                                       Memorize: NotMemorized,
                                       Access: ReadWrite},
                       'SaveDirectory': {Type: str,
                                         Description: "Directory to save the crystal files to",  # noqa
                                         Memorize: Memorized,
                                         Access: ReadWrite},
                       'AdjustAnglesToReflection': {Type: (float,),
                                                    Description: "Set the given angles to the reflection with given index",  # noqa
                                                    Access: ReadWrite},
                       'ReflectionAngles': {Type: ((float,), (float,)),
                                            Description: "Angles between reflections",  # noqa
                                            Access: ReadOnly},
                       'ModeParametersNames': {Type: (str,),  # TODO delete
                                               Description: "Name of the parameters of the current mode (if any)",  # noqa
                                               Access: ReadOnly},
                       'ModeParametersValues': {Type: (float,),  # TODO delete
                                                Description: "Value of the parameters of the current mode (if any)",  # noqa
                                                Access: ReadWrite},
                       'PsiRefH': {Type: float,
                                   Description: "x coordinate of the psi reference vector (-999 if not applicable)",  # noqa
                                   Memorize: MemorizedNoInit,
                                   Access: ReadWrite},
                       'PsiRefK': {Type: float,
                                   Description: "y coordinate of the psi reference vector (-999 if not applicable)",  # noqa
                                   Memorize: MemorizedNoInit,
                                   Access: ReadWrite},
                       'PsiRefL': {Type: float,
                                   Description: "z coordinate of the psi reference vector (-999 if not applicable)",  # noqa
                                   Memorize: MemorizedNoInit,
                                   Access: ReadWrite},
                       'MotorList': {Type: (str,),
                                     Description: "Name of the real motors",
                                     Access: ReadOnly},
                       'HKLPseudoMotorList': {Type: (str,),
                                              Description: "Name of the hkl pseudo motors",  # noqa
                                              Access: ReadOnly},
                       'MotorRoles': {Type: (str,),
                                          Description: "Name of the motor roles",  # noqa
                                          Access: ReadOnly},
                       'EnergyDevice': {Type: str,
                                        Description: "Name of the energy device to read the energy from",
                                        Memorize: Memorized,
                                        Access: ReadWrite},
                       'AutoEnergyUpdate': {Type: int,
                                            Description: "If 1 wavelength is read from EnergyDevice",
                                            Memorize: Memorized,
                                            Access: ReadWrite},
                       'ComputeHKL': {Type: (float,),
                                            Description: "Compute hkl for given angles",
                                            Memorize: NotMemorized,
                                            Access: ReadWrite},
    }

    axis_attributes = {
        'Mode': {
            Type: str,
            Memorize: MemorizedNoInit,
            Access: ReadWrite
        },
        'Modes': {
            Type: (str,),
            Access: ReadOnly
        },
        'ModeParameters': {
            Type: (str,),
            Description: "Name of the parameters of the current mode (if any)",  # noqa
            Access: ReadOnly
        },
        'ModeParametersValues': {
            Type: (float,),
            Description: "Value of the parameters of the current mode (if any)",  # noqa
            Access: ReadWrite
        },
    }

    def GetAxisExtraPar(self, axis, name):
        return getattr(self.axispar[axis - 1], name.lower())

    def SetAxisExtraPar(self, axis, name, value):
        setattr(self.axispar[axis - 1], name.lower(), value)

    MaxDevice = 1

    def __init__(self, inst, props, *args, **kwargs):
        """ Do the default init plus the specific diffractometer
        staff.
        @param properties of the controller
        """
        PseudoMotorController.__init__(self, inst, props, *args, **kwargs)

        # Comment out if memorized crystal (memcrystal)
        # self.first_crystal_set = 1
        self.samples = {}
        self.samples[DEFAULT_CRYSTAL] = self.sample = Hkl.Sample.new(DEFAULT_CRYSTAL)  # noqa

        lattice = self.sample.lattice_get()
        self._a = self._b = self._c = 1.5
        self._alpha = self._beta = self._gamma = 90.
        lattice.set(self._a, self._b, self._c,
                    self._alpha, self._beta, self._gamma, USER)
        self.sample.lattice_set(lattice)

        self.detector = Hkl.Detector.factory_new(Hkl.DetectorType(0))

        for key, factory in Hkl.factories().items():
            if key == self.DiffractometerType:
                self.geometry = factory.create_new_geometry()
                self.engines = factory.create_new_engine_list()

        self.nb_ph_axes = len(self.geometry.axis_names_get())

        self.engines.init(self.geometry, self.detector, self.sample)

        # Engine hkl -> it has not effect, it will be set the last value set
        # (stored in DB)
        self.engine = self.engines.engine_get_by_name("hkl")

        self.engine_list = []
        self.axispar = []
        for engine in self.engines.engines_get():
            self.engine_list.append(engine.name_get())
            for pseudo in engine.pseudo_axis_names_get():
                self.axispar.append(AxisPar(engine, pseudo))

        self.engine_list = tuple(self.engine_list)
        self.engines_conf = None  # defered because it does not work in the __init__

        self.trajectorylist = []

        # simulation part
        self.lastpseudopos = [0] * 3
        self.selected_trajectory = 0

        # Put to -1 the attributes that are actually commands
        self._affinecrystal = -1
        self._removereflection = -1
        self._deletecrystal = 'nothing'
        self._addcrystal = 'nothing'
        self._addreflection = [-1.]
        self._substitutereflection = [-1.]
        self._swapreflections01 = -1
        self._loadreflections = " "  # Only to create the member, the
        # value will be overwritten by
        # the one in stored in the
        # database
        self._savereflections = " "  # Only to create the member, the
        # value will be overwritten by
        # the one in stored in the
        # database
        self._loadcrystal = " "  # Only to create the member, the
        # value will be overwritten by the
        # one in stored in the database
        self._savecrystal = -1
        self._savedirectory = " "  # Only to create the member, the
        # value will be overwritten by the
        # one in stored in the database
        self._energydevice = " "  # Only to create the member, the
        # value will be overwritten by the
        # one in stored in the database
        self._autoenergyupdate = 0  # Only to create the member, the
        # value will be overwritten by the
        # one in stored in the database
        self._computehkl = []

        self.energy_device = None
        self.lambda_to_e = 12398.424  # Amstrong * eV

    def _solutions(self, values, curr_physical_position):
        # set all the motor min and max to restrain the solutions
        # with only valid positions.
        for role, current in zip(self.motor_roles, curr_physical_position):
            motor = self.GetMotor(role)
            axis = self.geometry.axis_get(role)
            axis.value_set(current, USER)
            try:
                config = PyTango.AttributeProxy(motor.get_full_name() + '/position').get_config()  # noqa
                mini, maxi = float(config.min_value), float(config.max_value)
                axis.min_max_set(mini, maxi, USER)
            except ValueError:
                pass
            self.geometry.axis_set(role, axis)

        # computation and select the expected solution
        return self.engine.pseudo_axis_values_set(values, USER)

    def CalcPhysical(self, axis, pseudo_pos, curr_physical_pos):
        return self.CalcAllPhysical(pseudo_pos, curr_physical_pos)[axis - 1]

    def CalcPseudo(self, axis, physical_pos, curr_pseudo_pos):
        return self.CalcAllPseudo(physical_pos, curr_pseudo_pos)[axis - 1]

    def CalcAllPhysical(self, pseudo_pos, curr_physical_pos):
        # TODO it should work with all the kind of engine ? or only
        # with the hkl engine ? What I understand from this is that
        # the pseudos values contain all the values from all the
        # engines. This is not generic at all. If I add new engines
        # this code should be fixed. worse the engine names is
        # hardcoded. for now I will just rewrite the code with the
        # same logic but with the new hkl API.

        engine_name = self.engine.name_get()
        values = None
        if engine_name == "hkl":
            values = [pseudo_pos[0], pseudo_pos[1], pseudo_pos[2]]
        elif self.nb_ph_axes == 4:  # noqa "E4CV", "E4CH", "SOLEIL MARS": hkl(3), psi(1), q(1); "K4CV": hkl(3), psi(1), q(1), eulerians(3); "SOLEIL SIXS MED1+2": hkl(3), q2(2), qper_qpar(2); "PETRA3 P23 4C": hkl(3), q2(2), qper_qpar(2), tth2(2), petra3_p23_4c_incidence(1), _petra3_p23_4c_emergence(1)
            if engine_name == "psi":
                values = [pseudo_pos[3]]
            elif engine_name == "q":
                values = [pseudo_pos[4]]
            elif engine_name == "eulerians":
                values = [pseudo_pos[5], pseudo_pos[6], pseudo_pos[7]]
            elif engine_name == "q2":
                values = [pseudo_pos[3], pseudo_pos[4]]
            elif engine_name == "qper_qpar":
                values = [pseudo_pos[5], pseudo_pos[6]]
            elif engine_name == "tth2":
                values = [pseudo_pos[7], pseudo_pos[8]]
            elif engine_name == "petra3_p23_4c_incidence":
                values = [pseudo_pos[9]]
            elif engine_name == "petra3_p23_4c_emergence":
                values = [pseudo_pos[10]]
        elif self.nb_ph_axes == 6:  # noqa "E6C", "SOLEIL SIXS MED2+2": hkl(3), psi(1), q2(2), qper_qpar(2); "K6C":  hkl(3), psi(1), q2(2), qper_qpar(2), eulerians(3); "PETRA3 P09 EH2": hkl(3), "PETRA3 P23 6C": hkl(3), psi(1), q2(2), qper_qpar(2), tth2(2), petra3_p23_6c_incidence(1), _petra3_p23_6c_emergence(1)
            if engine_name == "psi":
                values = [pseudo_pos[3]]
            elif engine_name == "q2":
                values = [pseudo_pos[4], pseudo_pos[5]]
            elif engine_name == "qper_qpar":
                values = [pseudo_pos[6], pseudo_pos[7]]
            elif engine_name == "eulerians":
                values = [pseudo_pos[8], pseudo_pos[9], pseudo_pos[10]]
            elif engine_name == "tth2":
                values = [pseudo_pos[8], pseudo_pos[9]]
            elif engine_name == "petra3_p23_6c_incidence":
                values = [pseudo_pos[10]]
            elif engine_name == "petra3_p23_6c_emergence":
                values = [pseudo_pos[11]]

        # getWavelength updates wavelength in the library in case automatic
        # energy update is set. Needed before computing trajectories.

        self.getWavelength()

        solutions = self._solutions(values, curr_physical_pos)
        if self.selected_trajectory > len(list(solutions.items())):
            self.selected_trajectory = len(list(solutions.items())) - 1
        for i, item in enumerate(solutions.items()):
            if i == self.selected_trajectory:
                angles = item.geometry_get().axis_values_get(USER)

        # TODO why replace this by a tuple ?
        return tuple(angles)

    def CalcAllPseudo(self, physical_pos, curr_pseudo_pos):
        # TODO howto avoid this nb_ph_axes, does the length of the
        # physical values are not equal to the expected len of the
        # geometry axes.

        # getWavelength updates wavelength in the library in case automatic
        # energy update is set. Needed before computing trajectories.

        self.getWavelength()

        # write the physical motor into the geometry
        self.geometry.axis_values_set(physical_pos[:self.nb_ph_axes], USER)
        self.engines.get()

        # extract all the pseudo axes values
        values = []
        for engine in self.engines.engines_get():
            values = values + engine.pseudo_axis_values_get(USER)

        return tuple(values)

    def getCrystal(self):
        return self.sample.name_get()

    def setCrystal(self, value):
        self.sample = self.samples[value]
        # Used this part and comment the above one out if memorized crystal (memcrystal)
        # if self.first_crystal_set == 0:
        #    if value in self.samples:
        #        self.sample = self.samples[value]
        # else:
        #    self.samples[value] = Hkl.Sample.new(
        #        value)  # By default a crystal is created with lattice parameters 1.54, 1.54, 1.54, 90., 90., 90.
        #    self.sample = self.samples[value]
        #    lattice = self.sample.lattice_get()
        #    lattice.set(self._a, self._b, self._c, self._alpha, self._beta, self._gamma, USER)
        #    self.sample.lattice_set(lattice)
        #    # For getting the UB matrix changing
        #    self.sample.lattice_set(lattice)

        #   self.first_crystal_set = 0
        self.engines.init(self.geometry, self.detector, self.sample)

    def setAffineCrystal(self, value):
        new_sample_name = self.sample.name_get() + " (affine)"
        if new_sample_name not in self.samples:
            sample = self.sample.copy()
            sample.name_set(new_sample_name)
            sample.affine()
            self.sample = self.samples[new_sample_name] = sample

    def getWavelength(self):
        if self._energydevice != " " and self._autoenergyupdate:
            try:
                if self.energy_device is None:
                    self.energy_device = PyTango.DeviceProxy(
                        self._energydevice)
                energy = self.energy_device.Position
                wavelength = self.lambda_to_e / energy
                self.setWavelength(wavelength)
            except:
                self._log.warning("Not able to get energy from energy device")
        return self.geometry.wavelength_get(USER)

    def setWavelength(self, value):
        self.geometry.wavelength_set(value, USER)

    def getEngineMode(self):
        return self.engine.current_mode_get()

    def setEngineMode(self, value):
        # TODO why not throw the exception when the mode name is
        # wrong. The hkl library return an usefull Exception for this.
        for mode in self.engine.modes_names_get():
            if value == mode:
                self.engine.current_mode_set(mode)

    def getEngineModeList(self):
        return self.engine.modes_names_get()

    def getHKLModeList(self):
        # TODO seems to me complicate... why this Hkl specific part.
        # It seems that this controleur mix all the engines and does
        # something special with the hkl one ???   neverthless I would
        # be possible to create a self.engines instead of recomputing
        # it all the time.
        for key, factory in Hkl.factories().items():
            if key == self.DiffractometerType:
                new_engines = factory.create_new_engine_list()
        new_engines.init(self.geometry, self.detector, self.sample)
        hkl_engine = new_engines.engine_get_by_name("hkl")
        return hkl_engine.modes_names_get()

    def getUBMatrix(self):
        UB = self.sample.UB_get()
        return [[UB.get(i, j) for j in range(3)] for i in range(3)]

    def getUx(self):
        return self.sample.ux_get().value_get(USER)

    def setUx(self, value):
        ux = self.sample.ux_get()
        ux.value_set(value, USER)
        self.sample.ux_set(ux)

    def getUy(self):
        return self.sample.uy_get().value_get(USER)

    def setUy(self, value):
        uy = self.sample.uy_get()
        uy.value_set(value, USER)
        self.sample.uy_set(uy)

    def getUz(self):
        return self.sample.uz_get().value_get(USER)

    def setUz(self, value):
        uz = self.sample.uz_get()
        uz.value_set(value, USER)
        self.sample.uz_set(uz)

    def setComputeUB(self, value):
        if len(value) < 2:
            return
        nb_reflections = len(self.sample.reflections_get())
        if value[0] >= nb_reflections or value[1] >= nb_reflections:
            return
        i = 0
        for ref in self.sample.reflections_get():
            if value[0] == i:
                ref1 = ref
            if value[1] == i:
                ref2 = ref
            i = i + 1
        self.sample.compute_UB_busing_levy(ref1, ref2)

    def getLatticeReciprocal(self):
        lattice = self.sample.lattice_get()
        reciprocal = lattice.copy()
        lattice.reciprocal(reciprocal)
        return reciprocal.get(USER)

    def getA(self):
        lattice = self.sample.lattice_get()
        a, _b, _c, _alpha, _beta, _gamma = lattice.get(USER)
        return a

    def setA(self, value):
        lattice = self.sample.lattice_get()
        a, b, c, alpha, beta, gamma = lattice.get(USER)
        lattice.set(value, b, c, alpha, beta, gamma, USER)
        self.sample.lattice_set(lattice)
        self._a = value

    def getB(self):
        lattice = self.sample.lattice_get()
        _a, b, _c, _alpha, _beta, _gamma = lattice.get(USER)
        return b

    def setB(self, value):
        lattice = self.sample.lattice_get()
        a, b, c, alpha, beta, gamma = lattice.get(USER)
        lattice.set(a, value, c, alpha, beta, gamma, USER)
        self.sample.lattice_set(lattice)
        self._b = value

    def getC(self):
        lattice = self.sample.lattice_get()
        _a, _b, c, _alpha, _beta, _gamma = lattice.get(USER)
        return c

    def setC(self, value):
        lattice = self.sample.lattice_get()
        a, b, c, alpha, beta, gamma = lattice.get(USER)
        lattice.set(a, b, value, alpha, beta, gamma, USER)
        self.sample.lattice_set(lattice)
        self._c = value

    def getAlpha(self):
        lattice = self.sample.lattice_get()
        _a, _b, _c, alpha, _beta, _gamma = lattice.get(USER)
        return alpha

    def setAlpha(self, value):
        lattice = self.sample.lattice_get()
        a, b, c, alpha, beta, gamma = lattice.get(USER)
        lattice.set(a, b, c, value, beta, gamma, USER)
        self.sample.lattice_set(lattice)
        self._alpha = value

    def getBeta(self):
        lattice = self.sample.lattice_get()
        _a, _b, _c, _alpha, beta, _gamma = lattice.get(USER)
        return beta

    def setBeta(self, value):
        lattice = self.sample.lattice_get()
        a, b, c, alpha, beta, gamma = lattice.get(USER)
        lattice.set(a, b, c, alpha, value, gamma, USER)
        self.sample.lattice_set(lattice)
        self._beta = value

    def getGamma(self):
        lattice = self.sample.lattice_get()
        _a, _b, _c, _alpha, _beta, gamma = lattice.get(USER)
        return gamma

    def setGamma(self, value):
        lattice = self.sample.lattice_get()
        a, b, c, alpha, beta, gamma = lattice.get(USER)
        lattice.set(a, b, c, alpha, beta, value, USER)
        self.sample.lattice_set(lattice)
        self._gamma = value

    def getAFit(self):
        apar = self.sample.lattice_get().a_get()
        return apar.fit_get()

    def setAFit(self, value):
        apar = self.sample.lattice_get().a_get()
        apar.fit_set(value)

    def getBFit(self):
        bpar = self.sample.lattice_get().b_get()
        return bpar.fit_get()

    def setBFit(self, value):
        bpar = self.sample.lattice_get().b_get()
        bpar.fit_set(value)

    def getCFit(self):
        cpar = self.sample.lattice_get().c_get()
        return cpar.fit_get()

    def setCFit(self, value):
        cpar = self.sample.lattice_get().c_get()
        cpar.fit_set(value)

    def getAlphaFit(self):
        alphapar = self.sample.lattice_get().alpha_get()
        return alphapar.fit_get()

    def setAlphaFit(self, value):
        alphapar = self.sample.lattice_get().alpha_get()
        alphapar.fit_set(value)

    def getBetaFit(self):
        betapar = self.sample.lattice_get().beta_get()
        return betapar.fit_get()

    def setBetaFit(self, value):
        betapar = self.sample.lattice_get().beta_get()
        betapar.fit_set(value)

    def getGammaFit(self):
        gammapar = self.sample.lattice_get().gamma_get()
        return gammapar.fit_get()

    def setGammaFit(self, value):
        gammapar = self.sample.lattice_get().gamma_get()
        gammapar.fit_set(value)

    def getComputeTrajectoriesSim(self):
        return self.lastpseudopos

    def setComputeTrajectoriesSim(self, values):
        # TODO the hkl library should return these informations
        # assert (len(values) == len(self.engine.pseudo_axis_names_get()),
        #         "Not the right number of parameters given (%d, %d expected)"
        #         " for the \"%s\" engine" %
        #         (len(values),
        #          len(self.engine.pseudo_axis_names_get()),
        #          self.engine.name_get())
        # getWavelength updates wavelength in the library in case automatic
        # energy update is set. Needed before computing trajectories.

        self.getWavelength()

        # Read current motor positions
        motor_position = []
        for i in range(0, self.nb_ph_axes):
            motor = self.GetMotor(i)
            motor_position.append(motor.get_position(cache=False).value)
        self.geometry.axis_values_set(motor_position, USER)

        curr_physical_pos = self.geometry.axis_values_get(USER)
        solutions = self._solutions(values, curr_physical_pos)
        self.trajectorylist = [item.geometry_get().axis_values_get(USER)
                               for item in list(solutions.items())]
        self.lastpseudos = tuple(values)

    def getTrajectoryList(self):
        return self.trajectorylist

    def getSelectedTrajectory(self):
        return self.selected_trajectory

    def setSelectedTrajectory(self, value):
        self.selected_trajectory = value

    def getEngine(self):
        return self.engine.name_get()

    def setEngine(self, value):
        self.engine = self.engines.engine_get_by_name(value)

    def getEngineList(self):
        return self.engine_list

    def getEnginesConf(self):
        if self.engines_conf is None:
            elements = []
            for engine in self.engines.engines_get():
                name = engine.name_get()
                motors = tuple([self.GetPseudoMotor(pseudo)
                                for pseudo in engine.pseudo_axis_names_get()])
                elements.append((name, motors))
            json_codec = CodecFactory().getCodec(format)
            f, conf = json_codec.encode(('', tuple(elements)))[1]
            self.engines_conf = conf

    def setAddCrystal(self, value):
        if value not in self.samples:
            self.samples[value] = Hkl.Sample.new(value)
            # value returned when the attribute is read
            self._addcrystal = value

    def getCrystalList(self):
        return list(self.samples)

    def setDeleteCrystal(self, value):
        if value in self.samples:
            self.samples.pop(value)
            # value returned when the attribute is read
            self._deletecrystal = value

    def setAddReflection(self, value):
        # Parameters: h, k, l, [affinement], angles are the current ones
        # Read current motor positions
        motor_position = []
        for i in range(0, self.nb_ph_axes):
            motor = self.GetMotor(i)
            motor_position.append(motor.get_position(cache=False).value)
        self.geometry.axis_values_set(motor_position, USER)
        newref = self.sample.add_reflection(
            self.geometry, self.detector, value[0], value[1], value[2])
        # Set affinement if given (by default reflections are created with
        # affinement = 1)
        if len(value) > 3:
            newref.flag_set(value[3])

    def setSubstituteReflection(self, value):
        # Parameters: index, h, k, l, [affinement], angles are the current ones
        # Read current reflections
        old_reflections = self.sample.reflections_get()
        nb_old_ref = len(old_reflections)
        hkla = []
        ihkla = 0
        angles = []
        for ref in old_reflections:
            hkla.append([])
            hkla[ihkla].append(ref.hkl_get()[0])
            hkla[ihkla].append(ref.hkl_get()[1])
            hkla[ihkla].append(ref.hkl_get()[2])
            hkla[ihkla].append(ref.flag_get())
            angles.append([])
            angles[ihkla].append(ihkla)
            for angle in ref.geometry_get().axis_values_get(USER):
                angles[ihkla].append(angle)
            ihkla = ihkla + 1

        # Remove reflections with index bigger than the inserted one
        if value[0] < nb_old_ref:
            for j in range(int(value[0]), nb_old_ref):
                # The index are shifted so we have to remove always de
                # first index
                self.setRemoveReflection(int(value[0]))

        # Check if the index is bigger than existing ones
        if value[0] < nb_old_ref:
            for i in range(0, nb_old_ref):
                if i < value[0]:
                    pass
                elif i == value[0]:
                    # add new reflection
                    self.setAddReflection(value[1:])
                elif i > value[0]:
                    self.setAddReflection(hkla[i])
                    self.setAdjustAnglesToReflection(angles[i])
        else:  # add the new one
            self.setAddReflection(value[1:])

    def setSwapReflections01(self, value):
        # Read current reflections
        reflections = self.sample.reflections_get()
        nb_ref = len(reflections)

        if nb_ref < 2:
            self._log.warning(
                "Only %d reflection(s) defined. Swap not possible" % (nb_ref,))
            return

        hkla = []
        ihkla = 0
        angles = []
        for ref in reflections:
            if ihkla < 2:
                hkla.append([])
                if ihkla == 1:
                    hkla[ihkla].append(0)
                else:
                    hkla[ihkla].append(1)
                hkla[ihkla].append(ref.hkl_get()[0])
                hkla[ihkla].append(ref.hkl_get()[1])
                hkla[ihkla].append(ref.hkl_get()[2])
                hkla[ihkla].append(ref.flag_get())
                angles.append([])
                # swap the index
                if ihkla == 1:
                    angles[ihkla].append(0)
                else:
                    angles[ihkla].append(1)
                for angle in ref.geometry_get().axis_values_get(USER):
                    angles[ihkla].append(angle)
            ihkla = ihkla + 1

        # Remove reflection 0
        self.setRemoveReflection(0)

        # Insert old ref 1 to 0

        self.setSubstituteReflection(hkla[1])
        self.setAdjustAnglesToReflection(angles[1])

        # Remove reflection 1

        self.setRemoveReflection(1)

        # Insert old ref 0 to 1

        self.setSubstituteReflection(hkla[0])
        self.setAdjustAnglesToReflection(angles[0])

    def getReflectionList(self):
        reflectionslist = []
        i = 0
        for ref in self.sample.reflections_get():
            reflectionslist.append([])
            reflectionslist[i].append(i)  # Reflection index
            reflectionslist[i].append(ref.hkl_get()[0])
            reflectionslist[i].append(ref.hkl_get()[1])
            reflectionslist[i].append(ref.hkl_get()[2])
            reflectionslist[i].append(0)  # Relevance
            reflectionslist[i].append(ref.flag_get())  # Affinement
            for value in ref.geometry_get().axis_values_get(USER):
                reflectionslist[i].append(value)
            i = i + 1
        return reflectionslist

    def setRemoveReflection(self, value):  # value: reflexion index
        i = 0
        for ref in self.sample.reflections_get():
            if i == value:
                self.sample.del_reflection(ref)
            i = i + 1

    # value: complete path of the file with the reflections to set
    def setLoadReflections(self, value):
        # Read the file
        with open(value, 'r') as f:
            self._loadreflections = value

            # Remove all reflections
            for reflection in self.sample.reflections_get():
                self.sample.del_reflection(reflection)

            # add one reflection per line (no check for now)
            # TODO it seems that the wavelength is missing in the file.
            for line in f:
                # the reflection line is structured like this
                # index 0 -> reflec. index;
                # index 1, 2, 3 -> hkl;
                # index 4 -> relevance
                # index 5 -> affinement;
                # last ones (2, 4 or 6) -> geometry axes values
                values = [float(v) for v in line.split(' ')]

                # create the reflection
                reflection = self.sample.add_reflection(self.geometry, self.detector,
                                                        values[1], values[2], values[3])
                # set the affinement
                reflection.flag_set(values[5])

                # set the axes values
                geometry = reflection.geometry_get()
                geometry.axis_values_set(values[6:], USER)
                reflection.geometry_set(geometry)

    def setLoadCrystal(self, value):
        """Load crystal information from a file. Ignore wavelength information.

        :param value: complete path of the file with the crystal to set
        :type value: :obj:`str`
        """
        # Read the file
        with open(value, 'r') as crystal_file:
            self._loadcrystal = value

            nb_ref = 0

            for line in crystal_file:
                line = line.replace("\n", "")
                if line.find("Crystal") != -1:
                    # Add crystal
                    line = line.replace(" ", "")
                    crystal = line.split("Crystal", 1)[1]
                    self.setAddCrystal(crystal)
                    # Set crystal
                    self.sample = self.samples[crystal]
                    self.engines.init(
                        self.geometry, self.detector, self.sample)
                    # Remove all reflections from crystal (there should not be
                    # any ... but just in case)
                    for ref in self.sample.reflections_get():
                        self.sample.del_reflection(ref)
                elif line.find("A") != -1 and line.find("B") != -1 and line.find("C") != -1:
                    par_line = line.split(" ")
                    avalue = float(par_line[1])
                    bvalue = float(par_line[3])
                    cvalue = float(par_line[5])
                    lattice = self.sample.lattice_get()
                    a, b, c, alpha, beta, gamma = lattice.get(USER)
                    lattice.set(avalue, bvalue, cvalue,
                                alpha, beta, gamma, USER)
                    # For getting the UB matrix changing
                    self.sample.lattice_set(lattice)
                    self._a = avalue
                    self._b = bvalue
                    self._c = cvalue
                elif line.find("Alpha") != -1 and line.find("Beta") != -1 and line.find("Gamma") != -1:
                    par_line = line.split(" ")
                    alphavalue = float(par_line[1])
                    betavalue = float(par_line[3])
                    gammavalue = float(par_line[5])
                    lattice = self.sample.lattice_get()
                    a, b, c, alpha, beta, gamma = lattice.get(USER)
                    lattice.set(a, b, c, alphavalue,
                                betavalue, gammavalue, USER)
                    # For getting the UB matrix changing
                    self.sample.lattice_set(lattice)
                    self._alpha = alphavalue
                    self._beta = betavalue
                    self._gamma = gammavalue
                elif line.find("Engine") != -1:
                    line = line.split(" ")
                    engine = line[1]
                    self.setEngine(engine)
                elif line.find("Mode") != -1:
                    line = line.split(" ")
                    mode = line[1]
                    self.setEngineMode(mode)
                elif line.find("PsiRef") != -1:
                    if line.find("PsiRef not available") == -1:
                        psiref_line = line.split(" ")
                        psirefh = float(psiref_line[1])
                        psirefk = float(psiref_line[2])
                        psirefl = float(psiref_line[3])
                        try:
                            self.setPsiRefH(psirefh)
                        except:
                            self._log.warning(
                                "PsiRefH not set. Psi not available in current mode")
                        try:
                            self.setPsiRefK(psirefk)
                        except:
                            self._log.warning(
                                "PsiRefK not set. Psi not available in current mode")
                        try:
                            self.setPsiRefL(psirefl)
                        except:
                            self._log.warning(
                                "PsiRefL not set. Psi not available in current mode")
                elif line.find("R0") != -1 or line.find("R1") != -1:
                    if line.find("R0") != -1:
                        line = line.split("R0 ")[1]
                    else:
                        line = line.split("R1 ")[1]
                    # Set reflections
                    ref_values = []
                    for value in line.split(' '):
                        try:
                            # index 0 -> reflec. index; index 1,2, 3 hkl; 4
                            # relevance; 5 affinement; last ones (2, 4 or 6)
                            # angles
                            ref_values.append(float(value))
                        except:
                            pass
                    # Set hkl values to the reflection
                    newref = self.sample.add_reflection(
                        self.geometry, self.detector, ref_values[1], ref_values[2], ref_values[3])
                    # Set affinement
                    newref.flag_set(ref_values[5])
                    # Adjust angles
                    new_angles = []
                    for i in range(6, len(ref_values)):
                        new_angles.append(ref_values[i])
                    geometry = newref.geometry_get()
                    geometry.axis_values_set(new_angles, USER)
                    newref.geometry_set(geometry)
                    nb_ref = nb_ref + 1
                elif line.find("SaveDirectory") != -1:
                    line = line.split(" ")
                    if self._savedirectory == " " or self._savedirectory == "":
                        self._savedirectory = line[1]
                elif line.find("AutoEnergyUpdate") != -1:
                    line = line.split(" ")
                    self._autoenergyupdate = int(line[1])

        if nb_ref > 1:
            values = [0, 1]
            self.setComputeUB(values)

    # value: directory, the file would be given by the name of the sample
    def setSaveReflections(self, value):
        complete_file_name = value + str(self.sample.name_get()) + ".ref"
        complete_file_name = complete_file_name.replace(' ', '')
        complete_file_name = complete_file_name.replace('(', '_')
        complete_file_name = complete_file_name.replace(')', '_')
        try:
            open(complete_file_name)
            new_file_name = complete_file_name + "_" + str(time.time())
            cmd = "mv " + str(complete_file_name) + " " + str(new_file_name)
            os.system(cmd)
        except:
            pass
        with open(complete_file_name, 'w') as ref_file:
            reflections = self.getReflectionList()
            for ref in reflections:
                ref_str = ""
                for val in ref:
                    ref_str = ref_str + str(val) + " "
                ref_str = ref_str[:-1]
                ref_str = ref_str + '\n'
                ref_file.write(ref_str)

    def setSaveCrystal(self, value):  # value: not used
        default_file_name = self._savedirectory + "/defaultcrystal.txt"
        crystal_file_name = self._savedirectory + "/" + self.sample.name_get() + ".txt"

        with open(default_file_name, 'w') as crys_file:
            # date

            date_str = "Created at " + time.strftime("%Y-%m-%d %H:%M") + "\n\n"
            crys_file.write(date_str)

            # diffractometer type

            difftype_str = ("DiffractometerType " + self.DiffractometerType +
                            "\n\n")
            crys_file.write(difftype_str)

            # write crystal name

            crystal_name = self.sample.name_get()
            crys_str = "Crystal    " + crystal_name + "\n\n"
            crys_file.write(crys_str)

            # write wavelength

            wavelength = self.geometry.wavelength_get(USER)
            wl_str = "Wavelength " + str(wavelength) + "\n\n"
            crys_file.write(wl_str)

            # write lattice parameters
            apar = self.sample.lattice_get().a_get()
            a = apar.value_get(USER)
            bpar = self.sample.lattice_get().b_get()
            b = bpar.value_get(USER)
            cpar = self.sample.lattice_get().c_get()
            c = cpar.value_get(USER)
            alphapar = self.sample.lattice_get().alpha_get()
            alpha = alphapar.value_get(USER)
            betapar = self.sample.lattice_get().beta_get()
            beta = betapar.value_get(USER)
            gammapar = self.sample.lattice_get().gamma_get()
            gamma = gammapar.value_get(USER)
            par_str = "A " + str(a) + " B " + str(b) + " C " + str(c) + "\n"
            crys_file.write(par_str)
            par_str = "Alpha " + str(alpha) + " Beta " + \
                str(beta) + " Gamma " + str(gamma) + "\n\n"
            crys_file.write(par_str)

            # write reflections
            reflections = self.getReflectionList()
            ref_in = 0
            for ref in reflections:
                ref_str = ""
                for val in ref:
                    ref_str = ref_str + str(val) + " "
                ref_str = ref_str[:-1]
                ref_str = "R" + str(ref_in) + " " + ref_str + '\n'
                if ref_in < 2:
                    crys_file.write(ref_str)
                ref_in = ref_in + 1
            if ref_in == 0:
                ref_str = "No reflections\n"
                crys_file.write(ref_str)
            crys_file.write("\n")

            # write engine
            engine_str = "Engine " + self.engine.name_get() + "\n\n"
            crys_file.write(engine_str)

            # write mode
            mode_str = "Mode " + self.engine.current_mode_get() + "\n\n"
            crys_file.write(mode_str)

            # write psiref (if available in mode)

            psirefh = self.getPsiRefH()
            psirefk = self.getPsiRefK()
            psirefl = self.getPsiRefL()
            if psirefh != -999 or psirefk != -999 or psirefl != -999:
                psi_str = "PsiRef " + \
                    str(psirefh) + " " + str(psirefk) + \
                    " " + str(psirefl) + "\n\n"
            else:
                psi_str = "PsiRef not available in current engine mode\n\n"

            crys_file.write(psi_str)

            # write autoenergyupdate value

            autoenergyupdate_str = "AutoEnergyUpdate " + \
                str(self._autoenergyupdate) + "\n\n"
            crys_file.write(autoenergyupdate_str)

            # Only for info in the file but not for loading:

            # write ub matrix
            for i in range(0, 3):
                ub_str = ""
                for j in range(0, 3):
                    #ub_str = ub_str + "U" + str(i) + str(j) + " " + str(self.sample.UB_get().get(i, j)) + " "
                    ub_str = ub_str + "U" + \
                        str(i) + str(j) + \
                        " %.3f" % self.sample.UB_get().get(i, j) + " "
                ub_str = ub_str + "\n"
                crys_file.write(ub_str)
            crys_file.write("\n")

            # write u vector
            u_str = "Ux " + str(self.sample.ux_get().value_get(USER)) + " Uy " + str(self.sample.uy_get(
            ).value_get(USER)) + " Uz " + str(self.sample.uz_get().value_get(USER)) + "\n\n"
            crys_file.write(u_str)

            # write directory where the file is saved

            dir_str = "SaveDirectory " + self._savedirectory + "\n"
            crys_file.write(dir_str)

        cmd = "cp " + str(default_file_name) + " " + str(crystal_file_name)
        os.system(cmd)

    def setSaveDirectory(self, value):
        self._savedirectory = value

    # value: reflexion index + new angles
    def setAdjustAnglesToReflection(self, value):
        ref_index = value[0]
        new_angles = value[1:]
        i = 0
        for ref in self.sample.reflections_get():
            if i == ref_index:
                geometry = ref.geometry_get()
                geometry.axis_values_set(new_angles, USER)
                ref.geometry_set(geometry)
            i = i + 1

    def getReflectionAngles(self):
        reflectionsangles = []
        i = -1
        for ref1 in self.sample.reflections_get():
            i = i + 1
            j = -1
            reflectionsangles.append([])
            for ref2 in self.sample.reflections_get():
                j = j + 1
                angle = 0
                if i < j:
                    angle = self.sample.get_reflection_measured_angle(
                        ref1, ref2)
                elif j < i:
                    angle = self.sample.get_reflection_theoretical_angle(
                        ref1, ref2)
                reflectionsangles[i].append(angle)
        return reflectionsangles

    def getModeParametersNames(self):
        return self.engine.parameters_names_get()

    def getModeParametersValues(self):
        return [p.value_get(USER)
                for p in [self.engine.parameter_get(n)
                          for n in self.engine.parameters_names_get()]]

    def setModeParametersValues(self, value):
        for parameter, v in zip(self.engine.parameters_names_get(), value):
            p = self.engine.parameter_get(parameter)
            p.value_set(v, USER)
            self.engine.parameter_set(parameter, p)

    def _getPsiRef(self, parameters):
        # TODO I do not understand this method. check that the
        # parameters names are ok. I changed one.
        value = -999
        for parameter in self.engine.parameters_names_get():
            if parameter in parameters:
                value = self.engine.parameter_get(parameter).value_get(USER)
        return value

    def getPsiRefH(self):
        return self._getPsiRef(["h1", "h2", "x"])

    def getPsiRefK(self):
        return self._getPsiRef(["k1", "k2", "y"])

    def getPsiRefL(self):
        return self._getPsiRef(["l1", "l2", "z"])

    def _setPsiRef(self, parameters, value):
        # TODO idem here :)
        value_set = False
        for parameter in self.engine.parameters_names_get():
            if parameter in parameters:
                p = self.engine.parameter_get(parameter)
                p.value_set(value, USER)
                self.engine.parameter_set(parameter, p)
                value_set = True
        if not value_set:
            raise Exception("psiref not available in this mode")

    def setPsiRefH(self, value):
        self._setPsiRef(["h1", "h2", "x"], value)

    def setPsiRefK(self, value):
        self._setPsiRef(["k1", "k2", "y"], value)

    def setPsiRefL(self, value):
        self._setPsiRef(["l1", "l2", "z"], value)

    def getMotorList(self):
        motor_names = []
        for i in range(0, self.nb_ph_axes):
            motor = self.GetMotor(i)
            mot_info = motor.name + "  (" + motor.full_name + ")"
            motor_names.append(mot_info)
        return motor_names

    def getHKLPseudoMotorList(self):
        hkl_pm_names = []
        for i in range(0, 3):
            motor = self.GetPseudoMotor(i)
            mot_info = motor.name + "  (" + motor.full_name + ")"
            hkl_pm_names.append(mot_info)
        return hkl_pm_names

    def getMotorRoles(self):
        roles_names = tuple(self.geometry.axis_names_get())
        return roles_names

    def setEnergyDevice(self, value):
        self._energydevice = value
        try:
            self.energy_device = PyTango.DeviceProxy(self._energydevice)
        except:
            self.energy_device = None
            self._log.warning("Not able to create proxy to energydevice")

    def setAutoEnergyUpdate(self, value):
        self._autoenergyupdate = value

    def setComputeHKL(self, value):

        # getWavelength updates wavelength in the library in case automatic
        # energy update is set. Needed before computing trajectories.

        self.getWavelength()

        # write the physical motor into the geometry
        if len(value) >= self.nb_ph_axes:
            self.geometry.axis_values_set(value[:self.nb_ph_axes], USER)
            self.engines.get()
        else:
            raise Exception("Not enough arguments. %d are need " %
                            (self.nb_ph_axes))

        # extract hkl  values
        values = []
        engine = self.engines.engine_get_by_name("hkl")
        values = engine.pseudo_axis_values_get(USER)

        self._computehkl = values


# 6C Diffractometers ####################


class Diffrac6Cp23(DiffracBasis):  # DiffractometerType: "PETRA3 P23 6C"

    """ The PseudoMotor controller for the diffractometer"""

    pseudo_motor_roles = "h", "k", "l", "psi", "q", "alpha", "tth2",
    "alpha_tth2", "incidence", "emergence"
    motor_roles = "omega_t", "mu", "omega", "chi", "phi", "gamma", "delta"

    def __init__(self, inst, props, *args, **kwargs):
        """ Do the default init plus the specific diffractometer
        staff.
        @param properties of the controller
        """

        DiffracBasis.__init__(self, inst, props, *args, **kwargs)

class Diffrac6C(DiffracBasis):  # DiffractometerType: "PETRA3 P09 EH2"

    """ The PseudoMotor controller for the diffractometer"""

    pseudo_motor_roles = "h", "k", "l"
    motor_roles = "mu", "omega", "chi", "phi", "delta", "gamma"

    def __init__(self, inst, props, *args, **kwargs):
        """ Do the default init plus the specific diffractometer
        staff.
        @param properties of the controller
        """

        DiffracBasis.__init__(self, inst, props, *args, **kwargs)


class DiffracE6C(DiffracBasis):

    """ The PseudoMotor controller for the diffractometer"""

    pseudo_motor_roles = "h", "k", "l", "psi", "q", "alpha", "qper", "qpar"
    motor_roles = "mu", "omega", "chi", "phi", "gamma", "delta"

    def __init__(self, inst, props, *args, **kwargs):
        """ Do the default init plus the specific diffractometer
        staff.
        @param properties of the controller
        """

        DiffracBasis.__init__(self, inst, props, *args, **kwargs)

# 4C Diffractometers ####################


class Diffrac4Cp23(DiffracBasis):  # DiffractometerType: "PETRA3 P23 4C"

    """ The PseudoMotor controller for the diffractometer"""

    pseudo_motor_roles = "h", "k", "l", "q", "alpha", "qper", "qpar",
    "tth2", "alpha_tth2", "incidence", "emergence"
    motor_roles = "omega_t", "mu", "gamma", "delta"

    def __init__(self, inst, props, *args, **kwargs):
        """ Do the default init plus the specific diffractometer
        staff.
        @param properties of the controller
        """

        DiffracBasis.__init__(self, inst, props, *args, **kwargs)

class DiffracE4C(DiffracBasis):

    """ The PseudoMotor controller for the diffractometer"""

    pseudo_motor_roles = "h", "k", "l", "psi", "q"
    motor_roles = "omega", "chi", "phi", "tth"

    def __init__(self, inst, props, *args, **kwargs):
        """ Do the default init plus the specific diffractometer
        staff.
        @param properties of the controller
        """

        DiffracBasis.__init__(self, inst, props, *args, **kwargs)


# wip Generic diffractometer

class Diffractometer(DiffracBasis):
    """ The PseudoMotor controller for the diffractometer"""

    def __init__(self, inst, props, *args, **kwargs):
        """ Do the default init plus the specific diffractometer
        staff.
        @param properties of the controller
        """
        DiffracBasis.__init__(self, inst, props, *args, **kwargs)

        factory = Hkl.factories()[self.DiffractometerType]
        self.geometry = factory.create_new_geometry()
        self.engines = factory.create_new_engine_list()

        # dynamically set the roles using the hkl library
        Diffractometer.motor_roles = tuple(self.geometry.axis_names_get())
        Diffractometer.pseudo_motor_roles = tuple(
            chain.from_iterable(
                [engine.pseudo_axis_names_get()
                 for engine in self.engines.engines_get()]))



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import time

import pytest
import unittest

from sardana import State
from sardana.pool.poolcountertimer import PoolCounterTimer
from sardana.pool.test import (FakePool, createPoolController,
                               createPoolCounterTimer, dummyCounterTimerConf01,
                               dummyPoolCTCtrlConf01)


class PoolCounterTimerTestCase(unittest.TestCase):
    """Unittest of PoolCounterTimer Class"""

    def setUp(self):
        """Create a Controller and a CounterTimer element"""
        pool = FakePool()

        pc = createPoolController(pool, dummyPoolCTCtrlConf01)
        self.pct = createPoolCounterTimer(pool, pc, dummyCounterTimerConf01)

    def test_init(self):
        """Verify that the created CounterTimer is a PoolCounterTimer
        instance."""
        msg = 'PoolCounterTimer constructor does not create ' +\
              'PoolCounterTimer instance'
        self.assertIsInstance(self.pct, PoolCounterTimer, msg)

    def test_acquisition(self):
        self.pct.integration_time = 0.1
        self.pct.start_acquisition()
        while self.pct.acquisition.is_running():
            time.sleep(0.01)
        msg = "wrong value after acquisition"
        self.assertEqual(self.pct.value.value, 0.1, msg)

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        self.pct = None


def StateOne_state(self, axis):
    return State.On


def StateOne_state_status(self, axis):
    return State.On, "Status"


@pytest.mark.parametrize("mock_StateOne", [StateOne_state,
                                           StateOne_state_status])
def test_state(monkeypatch, mock_StateOne):
    """Test variants of StateOne return value:
    - state
    - state, status
    """
    pool = FakePool()
    # when SEP19 gets implemented it should be possible to mock directly
    # the imported class
    DummyCounterTimerController = pool.ctrl_manager.getControllerClass(
        "DummyCounterTimerController")
    monkeypatch.setattr(DummyCounterTimerController, "StateOne",
                        mock_StateOne)
    ct_ctrl = createPoolController(pool, dummyPoolCTCtrlConf01)
    ct = createPoolCounterTimer(pool, ct_ctrl, dummyCounterTimerConf01)
    ct_ctrl.add_element(ct)
    pool.add_element(ct_ctrl)
    pool.add_element(ct)
    assert ct.state == State.On
    assert type(ct.status) == str



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import unittest
from sardana.pool.test import (FakePool, createPoolController,
                               dummyPoolCTCtrlConf01)
from sardana.pool.poolcontroller import PoolController


class PoolControllerTestCase(unittest.TestCase):
    """Unittest of PoolController Class"""

    def setUp(self):
        """Instantiate a fake Pool and create a Controller"""
        pool = FakePool()
        self.pc = createPoolController(pool, dummyPoolCTCtrlConf01)

    def test_init(self):
        """Verify that the created Controller is an instance of
        PoolController"""
        msg = 'PoolController constructor does not create ' +\
              'PoolController instance'
        self.assertIsInstance(self.pc, PoolController, msg)

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        self.pc = None



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool libray. It defines the base classes
for moveable elements"""

__all__ = ["PoolMoveable"]

__docformat__ = 'restructuredtext'


class PoolMoveable(object):

    def get_size(self):
        raise RuntimeError("Must be implemented in subclass")

    def calc_move(self, positions, ctrl_map, trust=False):
        raise RuntimeError("Must be implemented in subclass")

    def set_value(self, v, propagate=True):
        raise RuntimeError("Must be implemented in subclass")

    def get_value(self, cache=True):
        raise RuntimeError("Must be implemented in subclass")



##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import time
import ctypes
from threading import Condition


class CountLatch(object):
    """Synchronization primitive with the capacity to count and latch.
    Counting up latches, while reaching zero when counting down un-latches.

    .. note::
        The CountLatch class has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the module) may occur if
        deemed necessary by the core developers.
    """

    def __init__(self):
        self.count = 0
        self.condition = Condition()

    def count_up(self):
        """Count up."""
        self.condition.acquire()
        self.count += 1
        self.condition.release()

    def count_down(self, _=None):
        """Count down."""
        self.condition.acquire()
        self.count -= 1
        if self.count <= 0:
            self.condition.notifyAll()
        self.condition.release()

    def wait(self):
        """Wait until the counter reaches zero."""
        self.condition.acquire()
        while self.count > 0:
            self.condition.wait()
        self.condition.release()


_asyncexc = ctypes.pythonapi.PyThreadState_SetAsyncExc
# first define the async exception function args. This is
# absolutely necessary for 64 bits machines.
_asyncexc.argtypes = (ctypes.c_long, ctypes.py_object)


def raise_in_thread(exception, thread, logger=None):
    """Raise exception in a thread.

    Inspired on :meth:sardana.macroserver.macro.Macro.abort

    :param exception: Exception to be raised
    :param thread: thread in which raise the exception.
    """
    ret, i = 0, 0
    while ret != 1:
        th_id = ctypes.c_long(thread.ident)
        if logger:
            logger.debug("Sending AbortException to %s", thread.name)
        ret = _asyncexc(th_id, ctypes.py_object(exception))
        i += 1
        if ret == 0:
            # try again
            if i > 2:
                if logger:
                    logger.error("Failed to abort after three tries!")
                break
            time.sleep(0.1)
        if ret > 1:
            # if it returns a number greater than one, you're in trouble,
            # and you should call it again with exc=NULL to revert the
            # effect
            _asyncexc(th_id, None)
            if logger:
                logger.error("Failed to abort (unknown error code {})".
                             format(ret))
            break



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains the function to access sardana thread pool"""




__all__ = ["get_thread_pool"]

__docformat__ = 'restructuredtext'

import threading

from taurus.core.util.threadpool import ThreadPool, Worker


__thread_pool_lock = threading.Lock()
__thread_pool = None


class OmniWorker(Worker):

    def run(self):
        try:
            import tango
        except ImportError:
            Worker.run(self)
        # Tango is not thread safe when using threading.Thread. One must
        # use omni threads instead. This was confirmed for parallel
        # event subscriptions in PyTango#307. Use EnsureOmniThread introduced
        # in PyTango#327 whenever available.
        else:
            if hasattr(tango, "EnsureOmniThread"):
                with tango.EnsureOmniThread():
                    Worker.run(self)
            else:
                import taurus
                taurus.warning("Your Sardana system is affected by bug "
                               "tango-controls/pytango#307. Please use "
                               "PyTango with tango-controls/pytango#327.")
                Worker.run(self)


def get_thread_pool():
    """Returns the global pool of threads for Sardana

    :return: the global pool of threads object
    :rtype: taurus.core.util.ThreadPool"""

    global __thread_pool
    global __thread_pool_lock
    with __thread_pool_lock:
        if __thread_pool is None:
            # protect older versions of Taurus (without the worker_cls
            # argument) remove it whenever we bump Taurus dependency
            try:
                __thread_pool = ThreadPool(name="SardanaTP", Psize=10,
                                           worker_cls=OmniWorker)
            except TypeError:
                import taurus
                taurus.warning("Your Sardana system is affected by bug "
                               "tango-controls/pytango#307. Please use "
                               "Taurus with taurus-org/taurus#1081.")
                __thread_pool = ThreadPool(name="SardanaTP", Psize=10)

        return __thread_pool



##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

from sardana import State
from sardana.pool.controller import IORegisterController


class DummyIORController(IORegisterController):
    """This controller offers as many IORegisters as the user wants (up to
    1024)."""

    gender = "Simulation"
    model = "Basic"
    organization = "Sardana team"

    MaxDevice = 1024

    def __init__(self, inst, props, *args, **kwargs):
        IORegisterController.__init__(self, inst, props, *args, **kwargs)
        self.myvalue = 1

    def AddDevice(self, axis):
        self._log.debug('AddDevice %d' % axis)

    def DeleteDevice(self, axis):
        pass

    def StateOne(self, axis):
        return (State.On, "Device in On state")

    def ReadOne(self, axis):
        return self.myvalue

    def WriteOne(self, axis, value):
        self.myvalue = value

    def SendToCtrl(self, in_data):
        return ""



import unittest
from taurus.test import insertTest

from sardana.macroserver.macro import Type
from sardana.macroserver.msparameter import ParamDecoder, WrongParamType

from sardana.macroserver.mstypemanager import TypeManager


class FakeMacroServer(object):
    name = "FakeMacroServer"

macro_server = FakeMacroServer()
TYPE_MANAGER = TypeManager(macro_server)

params_def1 = [
    {
        "name": "param1",
        "type": Type.String,
        "default_value": None,
        "description": "param1 description"
    }
]

params_def2 = [
    {
        "name": "param1",
        "type": Type.Integer,
        "default_value": None,
        "description": "param1 description"
    }
]

params_def3 = [
    {
        "name": "param1",
        "type": [
            {
                "name": "subparam1",
                "type": Type.Integer,
                "default_value": None,
                "description": "subparam1 description",
            }
        ],
        "default_value": None,
        "description": "param1 description",
        "min": 0,
        "max": None
    }
]

params_raw1 = ["value1"]
params_raw2 = [[1]]
params_raw3 = []
expected_params1 = ["value1"]
expected_params2 = [[1]]
expected_params3 = []

doc1 = "Decode macro with one single parameter with correct value"
doc2 = "Decode macro with one single parameter with wrong value"
doc3 = "Decode macro with one simple repeat parameter with correct value"
doc4 = "Decode macro with one simple repeat parameter with min=0 and no values"


@insertTest(helper_name="decode", test_method_name="test_decode1",
            test_method_doc=doc1, params_def=params_def1,
            params_raw=params_raw1, expected_params=expected_params1)
@insertTest(helper_name="decode", test_method_name="test_decode2",
            test_method_doc=doc2, params_def=params_def2,
            params_raw=params_raw1, expected_exception=WrongParamType)
@insertTest(helper_name="decode", test_method_name="test_decode3",
            test_method_doc=doc3, params_def=params_def3,
            params_raw=params_raw1, expected_exception=WrongParamType)
@insertTest(helper_name="decode", test_method_name="test_decode4",
            test_method_doc=doc4, params_def=params_def3,
            params_raw=params_raw3, expected_params=expected_params3)
class TestParamDecoder(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.type_manager = TYPE_MANAGER
        # use this code to parameter types of sardana interfaces e.g. Motor
#         type_motor = self.type_manager.getTypeObj(Type.Motor)
#         type_motor.getObj = lambda name:name

    def decode(self, params_def, params_raw, expected_params=None,
               expected_exception=None):
        """Test decoding of macro parameters using ParamDecoder class by
        comparing the result with either of the expected_params or the
        expcted_exception result, not both of them.

        :param params_def: list(dict) where each dict represents one parameter
            definition (required keys: name, type, default_value, description
            and min & max in case of the repeat parameters)
        :param params_raw: list of parameter values or XML strucutre with
            parameter values
        :param expected_params: list of the expected parameter values after the
            decoding process
        :param expected_exception: expected exception class
        """
        if expected_params is None and expected_exception is None:
            raise ValueError("missing expected_params or expected_exception")
        if not expected_params is None and not expected_exception is None:
            raise ValueError("too many expected expected values")
        exception = None
        try:
            param_decoder = ParamDecoder(self.type_manager, params_def,
                                         params_raw)
        except Exception as e:
            exception = e
        if expected_params:
            exception_message = getattr(exception, "message", None)
            msg = "unexpected exception: %s (%s)" % (exception,
                                                     exception_message)
            self.assertIsNone(exception, msg)
            params = param_decoder.getParamList()
            msg = ("decoding result (%s) does not match with the expected"
                   " result (%s)" % (params, expected_params))
            self.assertListEqual(params, expected_params, msg)
        if expected_exception:
            msg = ("decoding exception type (%s) does not match with the"
                   " expected exception type (%s)" % (type(exception),
                                                      expected_exception))
            self.assertIsInstance(exception, expected_exception, msg)

    def tearDown(self):
        unittest.TestCase.tearDown(self)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################


__all__ = ["MacroServerManager"]

__docformat__ = 'restructuredtext'

import weakref

from taurus.core import ManagerState
from taurus.core.util.log import Logger


class MacroServerManager(Logger):
    """Base Class for macro server managers"""

    def __init__(self, macro_server):
        name = macro_server.name + "." + self.__class__.__name__
        self._state = ManagerState.UNINITIALIZED
        self._macro_server = weakref.ref(macro_server)
        self.call__init__(Logger, name)
        self.reInit()

    def reInit(self):
        self._state = ManagerState.INITED

    def cleanUp(self):
        self._state = ManagerState.CLEANED

    @property
    def macro_server(self):
        return self._macro_server()

    def is_initialized(self):
        return self._state == ManagerState.INITED

    def is_cleaned(self):
        return self._state == ManagerState.CLEANED



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This is the macro server scan data output recorder module"""

__all__ = ["FIO_FileRecorder", "NXscan_FileRecorder", "SPEC_FileRecorder"]

__docformat__ = 'restructuredtext'

import os
import time
import itertools
import re
import io
import weakref

import numpy

import PyTango

from sardana.taurus.core.tango.sardana import PlotType
from sardana.macroserver.macro import Type
from sardana.macroserver.scan.recorder import (BaseFileRecorder,
                                               BaseNAPI_FileRecorder,
                                               SaveModes)
from taurus.core.util.containers import chunks


class FIO_FileRecorder(BaseFileRecorder):
    """ Saves data to a file """

    formats = {'fio': '.fio'}

    def __init__(self, filename=None, macro=None, **pars):
        BaseFileRecorder.__init__(self)
        self.base_filename = filename
        self.macro = weakref.ref(macro) if macro else None
        self.db = PyTango.Database()
        if filename:
            self.setFileName(self.base_filename)

    def setFileName(self, filename):
        if self.fd is not None:
            self.fd.close()

        dirname = os.path.dirname(filename)

        if not os.path.isdir(dirname):
            try:
                os.makedirs(dirname)
            except:
                self.filename = None
                return
        self.currentlist = None
        #
        # construct the filename, e.g. : /dir/subdir/etcdir/prefix_00123.fio
        #
        tpl = filename.rpartition('.')
        try:  # For avoiding error when calling at __init__
            serial = self.recordlist.getEnvironValue('serialno')
            self.filename = "%s_%05d.%s" % (tpl[0], serial, tpl[2])
            #
            # in case we have MCAs, prepare the dir name
            #
            self.mcaDirName = "%s_%05d" % (tpl[0], serial)
        except:
            self.filename = "%s_%s.%s" % (tpl[0], "[ScanId]", tpl[2])

    def getFormat(self):
        return list(self.formats.keys())[0]

    def _startRecordList(self, recordlist):

        if self.base_filename is None:
            return

        self.setFileName(self.base_filename)

        envRec = recordlist.getEnviron()

        self.sampleTime = envRec['estimatedtime'] / \
            (envRec['total_scan_intervals'] + 1)
        # datetime object
        start_time = envRec['starttime']

        self.motorNames = envRec['ref_moveables']
        self.mcaNames = []
        self.ctNames = []
        for e in envRec['datadesc']:
            if len(e.shape) == 1:
                self.mcaNames.append(e.name)
            else:
                self.ctNames.append(e.name)
        #
        # we need the aliases for the column description
        #
        self.mcaAliases = []
        for mca in self.mcaNames:
            if mca.startswith("tango://"):
                mca = mca[8:]
            lst = mca.split("/")
            self.mcaAliases.append(self.db.get_alias("/".join(lst[1:])))

        # self.names = [ e.name for e in envRec['datadesc'] ]
        self.fd = open(self.filename, 'w')
        #
        # write the comment section of the header
        #
        self.fd.write("!\n! Comments\n!\n%%c\n %s\nuser %s Acquisition started at %s\n" %
                      (envRec['title'], envRec['user'], start_time.ctime()))
        self.fd.flush()
        #
        # write the parameter section, including the motor positions, if needed
        #
        self.fd.write("!\n! Parameter\n!\n%p\n")
        self.fd.flush()
        env = self.macro().getAllEnv()
        if ('FlagFioWriteMotorPositions' in env
                and env['FlagFioWriteMotorPositions']):
            all_motors = sorted(
                self.macro().findObjs('.*', type_class=Type.Motor))
            for mot in all_motors:
                pos = mot.getPosition(force=True)
                if pos is None:
                    record = "%s = nan\n" % (mot)
                else:
                    record = "%s = %g\n" % (mot, pos)

                self.fd.write(record)
            self.fd.flush()
        #
        # write the data section starting with the description of the columns
        #
        self.fd.write("!\n! Data\n!\n%d\n")
        self.fd.flush()
        i = 1
        for col in envRec['datadesc']:
            if col.name == 'point_nb':
                continue
            if col.name == 'timestamp':
                continue
            dType = 'FLOAT'
            if col.dtype == 'float64':
                dType = 'DOUBLE'
            outLine = " Col %d %s %s\n" % (i, col.label, dType)
            self.fd.write(outLine)
            i += 1
        #
        # 11.9.2012 timestamp to the end
        #
        outLine = " Col %d %s %s\n" % (i, 'timestamp', 'DOUBLE')
        self.fd.write(outLine)

        self.fd.flush()
        os.fsync(self.fd.fileno())

    def _writeRecord(self, record):
        if self.filename is None:
            return
        nan, ctNames, fd = float('nan'), self.ctNames, self.fd
        outstr = ''
        for c in ctNames:
            if c == "timestamp" or c == "point_nb":
                continue
            outstr += ' ' + str(record.data.get(c, nan))
        #
        # 11.9.2012 timestamp to the end
        #
        outstr += ' ' + str(record.data.get('timestamp', nan))
        outstr += '\n'

        fd.write(outstr)
        fd.flush()
        os.fsync(self.fd.fileno())

        if len(self.mcaNames) > 0:
            self._writeMcaFile(record)

    def _endRecordList(self, recordlist):
        if self.filename is None:
            return

        envRec = recordlist.getEnviron()
        end_time = envRec['endtime'].ctime()
        self.fd.write("! Acquisition ended at %s\n" % end_time)
        self.fd.flush()
        self.fd.close()

    def _writeMcaFile(self, record):
        if self.mcaDirName is None:
            return

        if not os.path.isdir(self.mcaDirName):
            try:
                os.makedirs(self.mcaDirName)
            except:
                self.mcaDirName = None
                return
        currDir = os.getenv('PWD')
        os.chdir(self.mcaDirName)

        serial = self.recordlist.getEnvironValue('serialno')
        if type(self.recordlist.getEnvironValue('ScanFile')).__name__ == 'list':
            scanFile = self.recordlist.getEnvironValue('ScanFile')[0]
        else:
            scanFile = self.recordlist.getEnvironValue('ScanFile')

        mcaFileName = "%s_%05d_mca_s%d.fio" % (
            scanFile.split('.')[0], serial, record.data['point_nb'] + 1)
        fd = open(mcaFileName, 'w')
        fd.write("!\n! Comments\n!\n%%c\n Position %g, Index %d \n" %
                 (record.data[self.motorNames[0]], record.data['point_nb']))
        fd.write("!\n! Parameter \n%%p\n Sample_time = %g \n" %
                 (self.sampleTime))
        self.fd.flush()

        col = 1
        fd.write("!\n! Data \n%d \n")
        for mca in self.mcaAliases:
            fd.write(" Col %d %s FLOAT \n" % (col, mca))
            col = col + 1

        if not record.data[self.mcaNames[0]] is None:
            # print "+++storage.py, recordno", record.recordno
            # print "+++storage.py, record.data", record.data
            # print "+++storage.py, len %d,  %s" % (len( record.data[ self.mcaNames[0]]), self.mcaNames[0])
            #
            # the MCA arrays me be of different size. the short ones are extended by zeros.
            #
            lMax = len(record.data[self.mcaNames[0]])
            for mca in self.mcaNames:
                if len(record.data[mca]) > lMax:
                    lMax = len(record.data[mca])

            for i in range(0, lMax):
                line = ""
                for mca in self.mcaNames:
                    if i > (len(record.data[mca]) - 1):
                        line = line + " 0"
                    else:
                        line = line + " " + str(record.data[mca][i])
                line = line + "\n"
                fd.write(line)

            fd.close()
        else:
            # print "+++storage.py, recordno", record.recordno, "data None"
            pass

        os.chdir(currDir)


class SPEC_FileRecorder(BaseFileRecorder):
    """ Saves data to a file """

    formats = {'Spec': '.spec'}
    supported_dtypes = ('float32', 'float64', 'int8',
                        'int16', 'int32', 'int64', 'uint8',
                        'uint16', 'uint32', 'uint64')

    def __init__(self, filename=None, macro=None, **pars):
        BaseFileRecorder.__init__(self)
        if filename:
            self.setFileName(filename)
        self._expectedlabels = pars.get('labels', None)

    def setLabels(self, labels):
        self._expectedlabels = labels

    def setFileName(self, filename):
        if self.fd is not None:
            self.fd.close()

        dirname = os.path.dirname(filename)

        if not os.path.isdir(dirname):
            try:
                os.makedirs(dirname)
            except:
                self.filename = None
                return
        self.filename = filename
        self.currentlist = None

    def getFormat(self):
        return list(self.formats.keys())[0]

    def _startRecordList(self, recordlist):
        '''Prepares and writes the scan header.'''
        if self.filename is None:
            return

        env = recordlist.getEnviron()

        # datetime object
        start_time = env['starttime']
        epoch = time.strftime("%a %b %d %H:%M:%S %Y", start_time.timetuple())
        serialno = env['serialno']

        # store names for performance reason
        labels = []
        names = []
        oned_labels = []
        oned_names = []
        oned_shape = 0
        _d_label_name = {}

        # REORGANIZE label and names
        for e in env['datadesc']:
            dims = len(e.shape)
            if dims >= 2:
                continue
            # substitute whitespaces by underscores and remove other
            # non-alphanumeric characters
            sanitizedlabel = "".join(x for x in e.label.replace(
                ' ', '_') if x.isalnum() or x == '_')
            if not dims or (dims == 1 and e.shape[0] == 1):
                labels.append(sanitizedlabel)
                names.append(e.name)
                _d_label_name[sanitizedlabel] = e.name
            else:
                oned_labels.append(sanitizedlabel)
                oned_names.append(e.name)
                oned_shape = e.shape[0]
        if self._expectedlabels:
            names = []
            labels = self._expectedlabels
            for label in labels:
                try:
                    names.append(_d_label_name[label])
                except KeyError:
                    msg = ('Try to generate a Spec File with an '
                           'non-existent label: %s' % (label))
                    raise Exception(msg)

        self.names = names
        self.oned_names = oned_names

        # prepare pre-scan snapshot
        snapshot_labels, snapshot_values = self._preparePreScanSnapshot(env)
        # format scan header
        data = {
            'serialno':  serialno,
            'title':     env['title'],
            'user':      env['user'],
            'epoch':     epoch,
            'starttime': start_time.ctime(),
            'nocols':    len(names),
            'labels':    '  '.join(labels)
        }
        # Compatibility with PyMca
        if os.path.exists(self.filename):
            header = '\n'
        else:
            header = ''
        header += '#S %(serialno)s %(title)s\n'
        header += '#U %(user)s\n'
        header += '#D %(epoch)s\n'
        header += '#C Acquisition started at %(starttime)s\n'
        # add a pre-scan snapshot (sep is two spaces for labels!!)
        header += self._prepareMultiLines('O', '  ', snapshot_labels)
        header += self._prepareMultiLines('P', ' ', snapshot_values)
        header += '#N %(nocols)s\n'
        if len(oned_labels) > 0:
            header += '#@MCA %sC\n' % oned_shape
            header += '#@CHANN %s 0 %s 1\n' % (oned_shape, oned_shape - 1)
            header += '#@MCA_NB %s\n' % len(oned_labels)
            for idx, oned_label in enumerate(oned_labels):
                header += '#@DET_%s %s\n' % (idx, oned_label)
        header += '#L %(labels)s\n'

        self.fd = io.open(self.filename, 'a', newline='\n')
        self.fd.write(str(header % data))
        self.fd.flush()
        os.fsync(self.fd.fileno())

    def _prepareMultiLines(self, character, sep, items_list):
        '''Translate list of lists of items into multiple line string

        :param character (string): each line will start #<character><line_nr>
        :sep: separator (string): separator to use between items
        :param items_list (list):list of lists of items

        :return multi_lines (string): string with all the items'''
        multi_lines = ''
        for nr, items in enumerate(items_list):
            start = '#%s%d ' % (character, nr)
            items_str = sep.join(map(str, items))
            end = '\n'
            line = start + items_str + end
            multi_lines += line
        return multi_lines

    def _preparePreScanSnapshot(self, env):
        '''Extract pre-scan snapshot, filters elements of shape different
        than scalar and split labels and values into chunks of 8 items.

        :param: env (dict) scan environment

        :return: labels, values (tuple<list,list>)
                 labels - list of chunks with 8 elements containing labels
                 values - list of chunks with 8 elements containing values
        '''
        # preScanSnapShot is a list o ColumnDesc objects
        pre_scan_snapshot = env.get('preScanSnapShot', [])
        labels = []
        values = []
        for column_desc in pre_scan_snapshot:
            shape = column_desc.shape  # shape is a tuple of dimensions
            label = column_desc.label
            dtype = column_desc.dtype
            pre_scan_value = column_desc.pre_scan_value
            # skip items
            # if label not in self._expectedlabels:
            #    continue
            # skip items with shape different than scalar
            if len(shape) > 0:
                self.info('Pre-scan snapshot of "%s" will not be stored.' +
                          ' Reason: value is non-scalar', label)
                continue
            if dtype not in self.supported_dtypes:
                self.info('Pre-scan snapshot of "%s" will not be stored.' +
                          ' Reason: type %s not supported', label, dtype)
                continue
            labels.append(label)
            values.append(pre_scan_value)
        # split labels in chunks o 8 items
        labels_chunks = list(chunks(labels, 8))
        values_chunks = list(chunks(values, 8))
        return labels_chunks, values_chunks

    def _writeRecord(self, record):
        if self.filename is None:
            return
        nan, names, fd = float('nan'), self.names, self.fd

        d = []
        for oned_name in self.oned_names:
            data = record.data.get(oned_name)
            # TODO: The method astype of numpy does not work properly on the
            # beamline, we found difference between the data saved on h5 and
            # spec. For that reason we implement the it by hand.
            #str_data = ' '.join(data.astype(str))
            str_data = ''
            if numpy.iterable(data):
                for i in data:
                    str_data += '%s ' % i
            else:
                str_data += '%s' % data
            outstr = '@A %s' % str_data
            outstr += '\n'
            fd.write(str(outstr))

        for c in names:
            data = record.data.get(c)
            if data is None:
                data = nan
            d.append(str(data))
        outstr = ' '.join(d)
        outstr += '\n'

        fd.write(str(outstr))

        fd.flush()
        os.fsync(self.fd.fileno())

    def _endRecordList(self, recordlist):
        if self.filename is None:
            return

        env = recordlist.getEnviron()
        end_time = env['endtime'].ctime()
        self.fd.write(str("#C Acquisition ended at %s\n" % end_time))
        self.fd.flush()
        self.fd.close()

    def _addCustomData(self, value, name, **kwargs):
        '''
        The custom data will be added as a comment line in the form::

        #C name : value

        ..note:: non-scalar values (or name/values containing end-of-line) will not be written
        '''
        if self.filename is None:
            self.info(
                'Custom data "%s" will not be stored in SPEC file. Reason: uninitialized file', name)
            return
        if numpy.ndim(value) > 0:  # ignore non-scalars
            self.info(
                'Custom data "%s" will not be stored in SPEC file. Reason: value is non-scalar', name)
            return
        v = str(value)
        if '\n' in v or '\n' in name:  # ignore if name or the string representation of the value contains end-of-line
            self.info(
                'Custom data "%s" will not be stored in SPEC file. Reason: unsupported format', name)
            return

        fileWasClosed = self.fd is None or self.fd.closed
        if fileWasClosed:
            try:
                self.fd = open(self.filename, 'a')
            except:
                self.info(
                    'Custom data "%s" will not be stored in SPEC file. Reason: cannot open file', name)
                return
        self.fd.write(str('#C %s : %s\n' % (name, v)))
        self.fd.flush()
        if fileWasClosed:
            self.fd.close()  # leave the file descriptor as found


class NXscan_FileRecorder(BaseNAPI_FileRecorder):
    """saves data to a nexus file that follows the NXscan application definition

        """

    def __init__(self, filename=None, macro=None, overwrite=False, **pars):
        BaseNAPI_FileRecorder.__init__(
            self, filename=filename, macro=macro, overwrite=overwrite, **pars)

    def _startRecordList(self, recordlist):
        nxs = self.nxs
        nxfilemode = self.getFormat()

        if self.filename is None:
            return

        self.currentlist = recordlist
        env = self.currentlist.getEnviron()
        serialno = env["serialno"]
        self._dataCompressionRank = env.get(
            "DataCompressionRank", self._dataCompressionRank)

        if not self.overwrite and os.path.exists(self.filename):
            nxfilemode = 'rw'
        self.fd = nxs.open(self.filename, nxfilemode)
        self.entryname = "entry%d" % serialno
        try:
            self.fd.makegroup(self.entryname, "NXentry")
        except self.nxs.NeXusError:
            entrynames = list(self.fd.getentries().keys())

            #==================================================================
            ##Warn and abort
            if self.entryname in entrynames:
                raise RuntimeError(('"%s" already exists in %s. To prevent data corruption the macro will be aborted.\n' % (self.entryname, self.filename) +
                                    'This is likely caused by a wrong ScanID\n' +
                                    'Possible workarounds:\n' +
                                    '  * first, try re-running this macro (the ScanID may be automatically corrected)\n'
                                    '  * if not, try changing ScanID with senv, or...\n' +
                                    '  * change the file name (%s will be in both files containing different data)\n' % self.entryname +
                                    '\nPlease report this problem.'))
            else:
                raise
            #==================================================================

            #==================================================================
            # Warn and continue writing to another entry
            # if self.entryname in entrynames:
            #    i = 2
            #    newname = "%s_%i"%(self.entryname,i)
            #    while(newname in entrynames):
            #        i +=1
            #        newname = "%s_%i"%(self.entryname,i)
            #    self.warning('"%s" already exists. Using "%s" instead. This may indicate a bug in %s',self.entryname, newname, self.macro.name)
            #    self.macro.warning('"%s" already exists. Using "%s" instead. \nThis may indicate a bug in %s. Please report it.',self.entryname, newname, self.macro.name)
            #    self.entryname = newname
            #    self.fd.makegroup(self.entryname,"NXentry")
            #==================================================================

        self.fd.opengroup(self.entryname, "NXentry")

        # adapt the datadesc to the NeXus requirements
        self.datadesc = []
        for dd in env['datadesc']:
            dd = dd.clone()
            dd.label = self.sanitizeName(dd.label)
            if dd.dtype == 'bool':
                dd.dtype = 'int8'
                self.debug('%s will be stored with type=%s', dd.name, dd.dtype)
            if dd.dtype in self.supported_dtypes:
                self.datadesc.append(dd)
            else:
                self.warning(
                    '%s will not be stored. Reason: type %s not supported', dd.name, dd.dtype)

        # make a dictionary out of env['instrumentlist'] (use fullnames -paths-
        # as keys)
        self.instrDict = {}
        for inst in env.get('instrumentlist', []):
            self.instrDict[inst.getFullName()] = inst
        if self.instrDict is {}:
            self.warning(
                "missing information on NEXUS structure. Nexus Tree won't be created")

        self.debug("starting new recording %d on file %s",
                   env['serialno'], self.filename)

        # populate the entry with some data
        # this is the Application Definition for NeXus Generic Scans
        self._writeData('definition', 'NXscan', 'char')
        import sardana.release
        program_name = "%s (%s)" % (
            sardana.release.name, self.__class__.__name__)
        self._writeData('program_name', program_name, 'char',
                        attrs={'version': sardana.release.version})
        # note: the type should be NX_DATE_TIME, but the nxs python api does
        # not recognize it
        self._writeData("start_time", env['starttime'].isoformat(), 'char')
        self.fd.putattr("epoch", time.mktime(env['starttime'].timetuple()))
        self._writeData("title", env['title'], 'char')
        self._writeData("entry_identifier", str(env['serialno']), 'char')
        # user data goes in a separate group following NX convention...
        self.fd.makegroup("user", "NXuser")
        self.fd.opengroup("user", "NXuser")
        self._writeData("name", env['user'], 'char')
        self.fd.closegroup()

        # prepare the "measurement" group
        self._createBranch("measurement:NXcollection")
        if self.savemode == SaveModes.Record:
            # create extensible datasets
            for dd in self.datadesc:
                self._makedata(dd.label, dd.dtype, [nxs.UNLIMITED] + list(dd.shape), chunks=[
                               1] + list(dd.shape))  # the first dimension is extensible
                if hasattr(dd, 'data_units'):
                    self.fd.opendata(dd.label)
                    self.fd.putattr('units', dd.data_units)
                    self.fd.closedata()

        else:
            # leave the creation of the datasets to _writeRecordList (when we
            # actually know the length of the data to write)
            pass

        self._createPreScanSnapshot(env)

        self.fd.flush()

    def _createPreScanSnapshot(self, env):
        # write the pre-scan snapshot in the
        # "measurement:NXcollection/pre_scan_snapshot:NXcollection" group
        self.preScanSnapShot = env.get('preScanSnapShot', [])
        self._createBranch(
            'measurement:NXcollection/pre_scan_snapshot:NXcollection')
        links = {}
        for dd in self.preScanSnapShot:  # desc is a ColumnDesc object
            label = self.sanitizeName(dd.label)
            dtype = dd.dtype
            pre_scan_value = dd.pre_scan_value
            if dd.dtype == 'bool':
                dtype = 'int8'
                pre_scan_value = numpy.int8(dd.pre_scan_value)
                self.debug(
                    'Pre-scan snapshot of %s will be stored with type=%s', dd.name, dtype)
            if dtype in self.supported_dtypes:
                # @todo: fallback shape is hardcoded!
                nid = self._writeData(
                    label, pre_scan_value, dtype, shape=dd.shape or (1,))
                links[label] = nid
            else:
                self.warning(
                    'Pre-scan snapshot of %s will not be stored. Reason: type %s not supported', dd.name, dtype)

        self.fd.closegroup()  # we are back at the measurement group

        measurement_entries = self.fd.getentries()
        for label, nid in list(links.items()):
            if label not in measurement_entries:
                self.fd.makelink(nid)

    def _writeRecord(self, record):
        if self.filename is None:
            return
        # most used variables in the loop
        fd, debug, warning = self.fd, self.debug, self.warning
        nparray, npshape = numpy.array, numpy.shape
        rec_data, rec_nb = record.data, record.recordno

        for dd in self.datadesc:
            if dd.name in record.data:
                data = rec_data[dd.name]
                fd.opendata(dd.label)

                if data is None:
                    data = numpy.zeros(dd.shape, dtype=dd.dtype)
                if not hasattr(data, 'shape'):
                    data = nparray([data], dtype=dd.dtype)
                elif dd.dtype != data.dtype.name:
                    debug('%s casted to %s (was %s)', dd.label, dd.dtype,
                          data.dtype.name)
                    data = data.astype(dd.dtype)

                slab_offset = [rec_nb] + [0] * len(dd.shape)
                shape = [1] + list(npshape(data))
                try:
                    fd.putslab(data, slab_offset, shape)
                except:
                    warning("Could not write <%s> with shape %s", data, shape)
                    raise

                # Note: the following 3 lines of code were substituted by the one above.
                # (now we trust the datadesc info instead of asking the nxs file each time)
                # shape,dtype=self.fd.getinfo()
                # shape[0]=1 #the shape of the record is of just 1 slab in the extensible dimension (first dim)
                # self.fd.putslab(record.data[lbl],[record.recordno]+[0]*(len(shape)-1),shape)
                fd.closedata()
            else:
                debug("missing data for label '%s'", dd.label)
        fd.flush()

    def _endRecordList(self, recordlist):

        if self.filename is None:
            return

        self._populateInstrumentInfo()
        self._createNXData()

        env = self.currentlist.getEnviron()
        self.fd.openpath("/%s:NXentry" % self.entryname)
        self._writeData("end_time", env['endtime'].isoformat(), 'char')
        self.fd.flush()
        self.debug("Finishing recording %d on file %s:",
                   env['serialno'], self.filename)
        # self.fd.show('.') #prints nexus file summary on stdout (only the
        # current entry)
        self.fd.close()
        self.currentlist = None

    def writeRecordList(self, recordlist):
        """Called when in BLOCK writing mode"""
        self._startRecordList(recordlist)
        for dd in self.datadesc:
            self._makedata(dd.label, dd.dtype, [
                           len(recordlist.records)] + list(dd.shape), chunks=[1] + list(dd.shape))
            self.fd.opendata(dd.label)
            try:
                # try creating a single block to write it at once
                block = numpy.array([r.data[dd.label]
                                     for r in recordlist.records], dtype=dd.dtype)
                # if dd.dtype !='char': block=numpy.array(block,dtype=dtype)
                # #char not supported anyway
                self.fd.putdata(block)
            except KeyError:
                # if not all the records contain this field, we cannot write it
                # as a block.. so do it record by record (but only this field!)
                for record in recordlist.records:
                    if dd.label in record.data:
                        self.fd.putslab(record.data[dd.label], [
                                        record.recordno] + [0] * len(dd.shape), [1] + list(dd.shape))
                    else:
                        self.debug(
                            "missing data for label '%s' in record %i", dd.label, record.recordno)
            self.fd.closedata()
        self._endRecordList(recordlist)

    def _populateInstrumentInfo(self):
        measurementpath = "/%s:NXentry/measurement:NXcollection" % self.entryname
        # create a link for each
        for dd in self.datadesc:
            if getattr(dd, 'instrument', None):  # we don't link if it is None or it is empty
                try:
                    datapath = "%s/%s" % (measurementpath, dd.label)
                    self.fd.openpath(datapath)
                    nid = self.fd.getdataID()
                    self._createBranch(dd.instrument)
                    self.fd.makelink(nid)
                except Exception as e:
                    self.warning(
                        "Could not create link to '%s' in '%s'. Reason: %s", datapath, dd.instrument, repr(e))

        for dd in self.preScanSnapShot:
            if getattr(dd, 'instrument', None):
                try:
                    label = self.sanitizeName(dd.label)
                    datapath = "%s/pre_scan_snapshot:NXcollection/%s" % (
                        measurementpath, label)
                    self.fd.openpath(datapath)
                    nid = self.fd.getdataID()
                    self._createBranch(dd.instrument)
                    self.fd.makelink(nid)
                except Exception as e:
                    self.warning(
                        "Could not create link to '%s' in '%s'. Reason: %s", datapath, dd.instrument, repr(e))

    def _createNXData(self):
        '''Creates groups of type NXdata by making links to the corresponding datasets
        '''
        # classify by type of plot:
        plots1d = {}
        plots1d_names = {}
        i = 1
        for dd in self.datadesc:
            ptype = getattr(dd, 'plot_type', PlotType.No)
            if ptype == PlotType.No:
                continue
            elif ptype == PlotType.Spectrum:
                # converting the list into a colon-separated string
                axes = ":".join(dd.plot_axes)
                if axes in plots1d:
                    plots1d[axes].append(dd)
                else:
                    plots1d[axes] = [dd]
                    # Note that datatesc ordering determines group name
                    # indexing
                    plots1d_names[axes] = 'plot_%i' % i
                    i += 1
            else:
                continue  # @todo: implement support for images and other

        # write the 1D NXdata group
        for axes, v in list(plots1d.items()):
            self.fd.openpath("/%s:NXentry" % (self.entryname))
            groupname = plots1d_names[axes]
            self.fd.makegroup(groupname, 'NXdata')
            # write the signals
            for i, dd in enumerate(v):
                src = "/%s:NXentry/measurement:NXcollection/%s" % (
                    self.entryname, dd.label)
                dst = "/%s:NXentry/%s:NXdata" % (self.entryname, groupname)
                self._nxln(src, dst)
                self.fd.opendata(dd.label)
                self.fd.putattr('signal', min(i + 1, 2))
                self.fd.putattr('axes', axes)
                self.fd.putattr('interpretation', 'spectrum')
            # write the axes
            for axis in axes.split(':'):
                src = "/%s:NXentry/measurement:NXcollection/%s" % (
                    self.entryname, axis)
                dst = "/%s:NXentry/%s:NXdata" % (self.entryname, groupname)
                try:
                    self._nxln(src, dst)
                except:
                    self.warning("cannot create link for '%s'. Skipping", axis)

    def _addCustomData(self, value, name, nxpath=None, dtype=None, **kwargs):
        '''
        apart from value and name, this recorder can use the following optional parameters:

        :param nxpath: (str) a nexus path (optionally using name:nxclass notation for
                       the group names). See the rules for automatic nxclass
                       resolution used by
                       :meth:`NXscan_FileRecorder._createBranch`.
                       If None given, it defaults to
                       nxpath='custom_data:NXcollection'

        :param dtype: name of data type (it is inferred from value if not given)

        '''
        if nxpath is None:
            nxpath = 'custom_data:NXcollection'
        if dtype is None:
            if numpy.isscalar(value):
                dtype = numpy.dtype(type(value)).name
                if numpy.issubdtype(dtype, str):
                    dtype = 'char'
                if dtype == 'bool':
                    value, dtype = int(value), 'int8'
            else:
                value = numpy.array(value)
                dtype = value.dtype.name

        if dtype not in self.supported_dtypes and dtype != 'char':
            self.warning(
                "cannot write '%s'. Reason: unsupported data type", name)
            return
        # open the file if necessary
        fileWasClosed = self.fd is None or not self.fd.isopen
        if fileWasClosed:
            if not self.overwrite and os.path.exists(self.filename):
                nxfilemode = 'rw'
            import nxs
            self.fd = nxs.open(self.filename, nxfilemode)
        # write the data
        self._createBranch(nxpath)
        try:
            self._writeData(name, value, dtype)
        except ValueError as e:
            msg = "Error writing %s. Reason: %s" % (name, str(e))
            self.warning(msg)
            self.macro.warning(msg)
        # leave the file as it was
        if fileWasClosed:
            self.fd.close()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Sardana, a Tango User Interface Library
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

__all__ = ("SarDemoEnv", "getElements", "getControllers", "getMotors",
           "getPseudoMotors", "getMoveables", "getCTs", "getZerods",
           "getOneds", "getTwods", "getIORs")

from taurus import Device
from sardana.taurus.core.tango.sardana import registerExtensions
from taurus.core.util.singleton import Singleton
from sardana import sardanacustomsettings
from sardana.sardanautils import is_number


class SarDemoEnv(Singleton):

    """Class to get _SAR_DEMO environment variable with cross checking with
    the MacroServer (given by :attr:`UNITTEST_DOOR_NAME`)
    """
    ready = False

    # when building docs, in RTD environment, init is never called - Singleton
    # (comming from taurus) is a mock
    def init(self, door_name=None):
        if door_name is None:
            door_name = getattr(sardanacustomsettings, 'UNITTEST_DOOR_NAME')

        # TODO: As a workaround: check with PyTango that the door is running
        # in case of exception raise a RuntimeError. Ideally it should be done
        # just using taurus
        try:
            import PyTango
            d = PyTango.DeviceProxy(door_name)
            # when building docs, in RTD environment, PyTango is a mock
            assert is_number(d.ping())
        except:
            raise RuntimeError("Door %s is not running" % door_name)

        registerExtensions()

        self.door = Device(door_name)
        self.ms = self.door.macro_server

        self.controllers = None
        self.cts = None
        self.motors = None
        self.pseudos = None
        self.zerods = None
        self.oneds = None
        self.twods = None

        try:
            self.env = self.ms.getEnvironment()['_SAR_DEMO']['elements'] + \
                list(self.ms.getEnvironment()['_SAR_DEMO']['controllers'])
        except KeyError:
            err = 'sar_demo has not been executed (or door %s not ready)' % \
                  door_name
            raise RuntimeError(err)
        self.ready = True

    def getElements(self, elem_type='all'):
        """Return the name of sardemo element(s) of given elem type

        :param elem_type: (str) type of elemnts to return (all by default)

        :return: (list<str>)
        """
        if not self.ready:
            raise RuntimeError(
                'SarDemoEnv instantiation had previously failed')
        if elem_type.lower() == 'all':
            return self.env
        if elem_type.lower() == 'moveable':
            return self.getElements('motor') + self.getElements('pseudomotor')
        ms_elems = self.ms.getElementNamesOfType(elem_type)
        elems = [e for e in ms_elems if e is not None and e in self.env]
        return elems

    def getMoveables(self):
        """Return the name of moveable(s) defined by SarDemo

        :return: (list<str>)
        """
        return self.getMotors() + self.getPseudoMotors()

    def getControllers(self):
        """Return the name of controllers(s) defined by SarDemo

        :return: (list<str>)
        """
        if not self.controllers:
            self.controllers = self.getElements('controller')
        return self.controllers

    def getCTs(self):
        """Return the name of counter timer exp channel(s) defined by SarDemo

        :return: (list<str>)
        """
        if not self.cts:
            self.cts = self.getElements('ctexpchannel')
        return self.cts

    def getMotors(self):
        """Return the name of motor(s) defined by SarDemo

        :return: (list<str>)
        """
        if not self.motors:
            self.motors = self.getElements('motor')
        return self.motors

    def getPseudoMotors(self):
        """Return the name of pseudomotor(s) defined by SarDemo

        :return: (list<str>)
        """
        if not self.pseudos:
            self.pseudos = self.getElements('pseudomotor')
        return self.pseudos

    def getZerods(self):
        """Return the name of zerod exp channel(s) defined by SarDemo

        :return: (list<str>)
        """
        if not self.zerods:
            self.zerods = self.getElements('zerodexpchannel')
        return self.zerods

    def getOneds(self):
        """Return the name of one exp channel(s) defined by SarDemo

        :return: (list<str>)
        """
        if not self.oneds:
            self.oneds = self.getElements('onedexpchannel')
        return self.oneds

    def getTwods(self):
        """Return the name of two exp channel(s) defined by SarDemo

        :return: (list<str>)
        """
        if not self.twods:
            self.twods = self.getElements('twodexpchannel')
        return self.twods

    def changeDoor(self, door_name):
        """Change the door name and reset all lists
        """
        self.__init__(door_name)


def getElements(elem_type="all", fallback_name="element_not_defined",
                fallback_elements_len=5):
    if fallback_name is None:
        fallback_name = elem_type + "_not_defined"
    try:
        elements = SarDemoEnv().getElements(elem_type)
    except RuntimeError:
        import taurus
        from sardana import sardanacustomsettings
        door_name = getattr(sardanacustomsettings, 'UNITTEST_DOOR_NAME',
                            'UNDEFINED')
        taurus.warning("The door %s is not running. " % (door_name) +
                       "Ignore this message if you are building the documentation.")
        elements = [fallback_name] * fallback_elements_len
    except Exception as e:
        import taurus
        taurus.debug(e)
        taurus.warning("It was not possible to retrieve the element. " +
                       "Ignore this message if you are building the documentation.")
        elements = [fallback_name] * fallback_elements_len
    return elements


def getControllers():
    return getElements(elem_type="controller")


def getCTs():
    return getElements(elem_type="ctexpchannel")


def getMotors():
    return getElements(elem_type="motor")


def getPseudoMotors():
    return getElements(elem_type="pseudomotor")


def getMoveables():
    return getElements(elem_type="moveable")


def getZerods():
    return getElements(elem_type="zerodexpchannel")


def getOneds():
    return getElements(elem_type="onedexpchannel")


def getTwods():
    return getElements(elem_type="twodexpchannel")


def getIORs():
    return getElements(elem_type="ioregister")


if __name__ == '__main__':
    s = SarDemoEnv()
    print(s.env)
    print(s.getControllers())
    print(s.getCTs())
    print(s.getMotors())
    print(s.getPseudoMotors())
    print(s.getZerods())
    print(s.getOneds())
    print(s.getTwods())
    print(s.getElements('Moveable'))
    print(s.getMoveables())
    print(s.getElements())



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool library. It defines the base classes
for Pool object"""

__all__ = ["PoolBaseObject"]

__docformat__ = 'restructuredtext'

from sardana.sardanabase import SardanaBaseObject


class PoolBaseObject(SardanaBaseObject):
    """The Pool most abstract object."""

    def __init__(self, **kwargs):
        kwargs['manager'] = kwargs.pop('pool')
        SardanaBaseObject.__init__(self, **kwargs)

    def get_pool(self):
        """Return the :class:`sardana.pool.pool.Pool` which *owns* this pool
        object.

        :return: the pool which *owns* this pool object.
        :rtype: :class:`sardana.pool.pool.Pool`"""
        return self.get_manager()

    def serialize(self, *args, **kwargs):
        kwargs = SardanaBaseObject.serialize(self, *args, **kwargs)
        kwargs['pool'] = self.pool.name
        return kwargs

    pool = property(get_pool,
                    doc="reference to the :class:`sardana.pool.pool.Pool`")



#!/usr/bin/env python

import sys
import time
import PyTango

from upgrade_utils import *

if __name__ == "__main__":
    try:
        import PyQt4
        import upgradeGUI
        upgradeGUI.main()
    except Exception, e:
        import upgradeCLI
        upgradeCLI.main()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import sys

import sardana
from taurus.external.qt import Qt
from taurus.qt.qtgui.container import TaurusWidget
from .reflectionslist import ReflectionsList
from .reflectionseditor import ReflectionsEditor
from taurus.qt.qtgui.base import TaurusBaseWidget

from taurus.external.qt import QtCore, QtGui

from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtgui.input import TaurusValueLineEdit
import taurus.core

import taurus.core.util.argparse
import taurus.qt.qtgui.application
from taurus.qt.qtgui.util.ui import UILoadable

global flag_update
flag_update = 0

__docformat__ = 'restructuredtext'


class PrivateComboBox(Qt.QComboBox, TaurusBaseWidget):
    """ComboBox"""

    def __init__(self, parent=None):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QComboBox, parent)
        self.call__init__(TaurusBaseWidget, name)
        self.setSizeAdjustPolicy(Qt.QComboBox.AdjustToContentsOnFirstShow)
        self.setToolTip("Choose an item ...")
        QtCore.QMetaObject.connectSlotsByName(self)

    def loadItems(self, items):
        all_items = [self.itemText(i) for i in range(self.count())]
        for crys in items:
            if crys not in all_items:
                self.addItem(crys)


@UILoadable(with_ui="_ui")
class UBMatrixBase(TaurusWidget):

    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent, designMode=designMode)

        self.loadUi(filename="ubmatrix.ui")

        self._ui.UpdateButton.clicked.connect(self.update_values)
        self._ui.ComputeUButton.clicked.connect(self.compute_ub)
        self._ui.ReflectionsListButton.clicked.connect(
            self.reflections_list_window)
        self._ui.EditReflectionsButton.clicked.connect(
            self.edit_reflections_window)
        self._ui.AffineButton.clicked.connect(self.affine)
        self._ui.AddCrystalButton.clicked.connect(self.add_select_crystal)
#        self._ui.alattice_value.textEdited.connect(self.on_alattice_value_textEdited)
#       Funciona con puro QEditValue pero no con TaurusQEdit ...

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusWidget.getQtDesignerPluginInfo()
        ret['module'] = 'ubmatrix'
        ret['group'] = 'Taurus Containers'
        ret['container'] = ':/designer/frame.png'
        ret['container'] = False
        return ret

    def setModel(self, model):

        self.model = model

        if model is not None:
            self.device = taurus.Device(model)

        self.update_values()

        uxmodel = model + "/ux"
        self._ui.taurusuxvalue.setModel(uxmodel)
        self._ui.taurusuxeditvalue.setModel(uxmodel)
        uymodel = model + "/uy"
        self._ui.taurusuyvalue.setModel(uymodel)
        self._ui.taurusuyeditvalue.setModel(uymodel)
        uzmodel = model + "/uz"
        self._ui.taurusuzvalue.setModel(uzmodel)
        self._ui.taurusuzeditvalue.setModel(uzmodel)
        amodel = model + "/a"
        self._ui.taurusalatticevalue.setModel(amodel)
        self._ui.taurusalatticeeditvalue.setModel(amodel)
        bmodel = model + "/b"
        self._ui.taurusblatticevalue.setModel(bmodel)
        self._ui.taurusblatticeeditvalue.setModel(bmodel)
        cmodel = model + "/c"
        self._ui.taurusclatticevalue.setModel(cmodel)
        self._ui.taurusclatticeeditvalue.setModel(cmodel)
        alphamodel = model + "/alpha"
        self._ui.taurusalphalatticevalue.setModel(alphamodel)
        self._ui.taurusalphalatticeeditvalue.setModel(alphamodel)
        betamodel = model + "/beta"
        self._ui.taurusbetalatticevalue.setModel(betamodel)
        self._ui.taurusbetalatticeeditvalue.setModel(betamodel)
        gammamodel = model + "/gamma"
        self._ui.taurusgammalatticevalue.setModel(gammamodel)
        self._ui.taurusgammalatticeeditvalue.setModel(gammamodel)
        psirefhmodel = model + "/psirefh"
        self._ui.taurusvalue_psirefh.setModel(psirefhmodel)
        self._ui.tauruseditvalue_psirefh.setModel(psirefhmodel)
        psirefkmodel = model + "/psirefk"
        self._ui.taurusvalue_psirefk.setModel(psirefkmodel)
        self._ui.tauruseditvalue_psirefk.setModel(psirefkmodel)
        psireflmodel = model + "/psirefl"
        self._ui.taurusvalue_psirefl.setModel(psireflmodel)
        self._ui.tauruseditvalue_psirefl.setModel(psireflmodel)
        wavelengthmodel = model + "/wavelength"
        self._ui.tauruswavelengthvalue.setModel(wavelengthmodel)

        # Set model to engine and modes

        enginemodel = model + '/engine'
        self._ui.taurusLabelEngine.setModel(enginemodel)

        self.enginescombobox = PrivateComboBox(self)
        self.enginescombobox.setGeometry(QtCore.QRect(130, 460, 221, 27))
        self.enginescombobox.setObjectName("engineslist")

        self.enginescombobox.loadItems(self.device.enginelist)

        self.enginescombobox.currentIndexChanged['QString'].connect(
            self.onEngineChanged)

        enginemodemodel = model + '/enginemode'
        self._ui.taurusLabelEngineMode.setModel(enginemodemodel)

        self.enginemodescombobox = PrivateComboBox(self)
        self.enginemodescombobox.setGeometry(QtCore.QRect(130, 500, 221, 27))
        self.enginemodescombobox.setObjectName("enginemodeslist")

        self.enginemodescombobox.loadItems(self.device.enginemodelist)

        self.enginemodescombobox.currentIndexChanged['QString'].connect(
            self.onModeChanged)

        # Set model to crystal

        crystalmodel = model + '/crystal'
        self._ui.taurusLabelCrystal.setModel(crystalmodel)

        self.crystalscombobox = PrivateComboBox(self)
        self.crystalscombobox.setGeometry(QtCore.QRect(130, 540, 221, 27))
        self.crystalscombobox.setObjectName("crystallist")

        self.crystalscombobox.loadItems(self.device.crystallist)

        self.crystalscombobox.currentIndexChanged['QString'].connect(
            self.onCrystalChanged)

    Qt.pyqtSlot('QString')
    def onEngineChanged(self, enginename):
        self.device.write_attribute("engine", str(enginename))

    Qt.pyqtSlot('QString')
    def onModeChanged(self, modename):
        self.device.write_attribute("enginemode", str(modename))

    Qt.pyqtSlot('QString')
    def onCrystalChanged(self, crystalname):
        if str(crystalname) != "":
            self.device.write_attribute("crystal", str(crystalname))

    def update_values(self):
        ub_values = self.device.ubmatrix
        self._ui.taurusub11value.setValue(ub_values[0][0])
        self._ui.taurusub12value.setValue(ub_values[0][1])
        self._ui.taurusub13value.setValue(ub_values[0][2])
        self._ui.taurusub21value.setValue(ub_values[1][0])
        self._ui.taurusub22value.setValue(ub_values[1][1])
        self._ui.taurusub23value.setValue(ub_values[1][2])
        self._ui.taurusub31value.setValue(ub_values[2][0])
        self._ui.taurusub32value.setValue(ub_values[2][1])
        self._ui.taurusub33value.setValue(ub_values[2][2])

        global flag_update
        if flag_update:
            all_items = [self.crystalscombobox.itemText(
                i) for i in range(self.crystalscombobox.count())]
            for crys in self.device.crystallist:
                if crys not in all_items:
                    self.crystalscombobox.addItem(crys)
            for i in range(self.crystalscombobox.count()):
                if self.crystalscombobox.itemText(i) not in self.device.crystallist:
                    self.crystalscombobox.removeItem(i)
        flag_update = 1

    def compute_ub(self):
        index = [0, 1]

        self.device.write_attribute("computeub", index)
        self.update_values()

    def reflections_list_window(self):

        reflections = self.device.reflectionlist

        nb_ref = 0
        xindex = 20
        xh = 70
        xk = 150
        xl = 230
        xrelevance = 330
        xaffinement = 380
        xangle1 = 430
        xangle2 = 510
        xangle3 = 590
        xangle4 = 670
        xangle5 = 750
        xangle6 = 830
        yhkl = 100
        w = ReflectionsList()

        self.taurusValueIndex = []
        self.taurusValueH = []
        self.taurusValueK = []
        self.taurusValueL = []
        self.taurusValueRelevance = []
        self.taurusValueAffinement = []
        self.taurusValueAngle1 = []
        self.taurusValueAngle2 = []
        self.taurusValueAngle3 = []
        self.taurusValueAngle4 = []
        self.taurusValueAngle5 = []
        self.taurusValueAngle6 = []

        if reflections is not None:
            for ref in reflections:
                if nb_ref == 0:
                    self.rl_label1_7 = QtGui.QLabel(w)
                    self.rl_label1_7.setGeometry(
                        QtCore.QRect(xangle1 + 20, 70, 51, 20))
                    self.rl_label1_7.setObjectName("rl_label1_7")
                #               self.testlabel.setLayoutDirection(QtCore.Qt.RightToLeft)
                    self.rl_label1_8 = QtGui.QLabel(w)
                    self.rl_label1_8.setGeometry(
                        QtCore.QRect(xangle2 + 20, 70, 71, 20))
                    self.rl_label1_8.setObjectName("rl_label1_8")
                #               self.testlabel.setLayoutDirection(QtCore.Qt.RightToLeft)
                    self.rl_label1_9 = QtGui.QLabel(w)
                    self.rl_label1_9.setGeometry(
                        QtCore.QRect(xangle3 + 20, 70, 41, 20))
                    self.rl_label1_9.setObjectName("rl_label1_9")
                    self.rl_label1_10 = QtGui.QLabel(w)
                    self.rl_label1_10.setGeometry(
                        QtCore.QRect(xangle4 + 20, 70, 41, 20))
                    self.rl_label1_10.setObjectName("rl_label1_10")
                    try:
                        self.angle_names = self.device.motorroles
                    except:  # Only for compatibility
                        if self.nb_motors == 4:
                            self.angles_names.append("omega")
                            self.angles_names.append("chi")
                            self.angles_names.append("phi")
                            self.angles_names.append("theta")
                        elif self.nb_motors == 6:
                            self.angles_names.append("mu")
                            self.angles_names.append("th")
                            self.angles_names.append("chi")
                            self.angles_names.append("phi")
                            self.angles_names.append("gamma")
                            self.angles_names.append("delta")
                # 4circles diffractometer
                    if len(ref) == 10:
                        self.rl_label1_7.setText(QtGui.QApplication.translate(
                            "Form", self.angle_names[0], None))
                        self.rl_label1_8.setText(QtGui.QApplication.translate(
                            "Form", self.angle_names[1], None))
                        self.rl_label1_9.setText(QtGui.QApplication.translate(
                            "Form", self.angle_names[2], None))
                        self.rl_label1_10.setText(QtGui.QApplication.translate(
                            "Form", self.angle_names[3], None))
                # 6 circles diffractometer
                    elif len(ref) == 12:
                        self.rl_label1_11 = QtGui.QLabel(w)
                        self.rl_label1_11.setGeometry(
                            QtCore.QRect(xangle5 + 20, 70, 71, 20))
                        self.rl_label1_11.setObjectName("rl_label1_11")
                        self.rl_label1_12 = QtGui.QLabel(w)
                        self.rl_label1_12.setGeometry(
                            QtCore.QRect(xangle6 + 20, 70, 41, 20))
                        self.rl_label1_12.setObjectName("rl_label1_12")
                        self.rl_label1_7.setText(QtGui.QApplication.translate(
                            "Form", self.angle_names[0], None))
                        self.rl_label1_8.setText(QtGui.QApplication.translate(
                            "Form", self.angle_names[1], None))
                        self.rl_label1_9.setText(QtGui.QApplication.translate(
                            "Form", self.angle_names[2], None))
                        self.rl_label1_10.setText(QtGui.QApplication.translate(
                            "Form", self.angle_names[3], None))
                        self.rl_label1_11.setText(QtGui.QApplication.translate(
                            "Form", self.angle_names[4], None))
                        self.rl_label1_12.setText(QtGui.QApplication.translate(
                            "Form", self.angle_names[5], None))

                self.taurusValueIndex.append(TaurusValueLineEdit(w))
                self.taurusValueIndex[nb_ref].setGeometry(
                    QtCore.QRect(xindex, 100 + 30 * (nb_ref), 41, 27))
                self.taurusValueIndex[nb_ref].setReadOnly(True)
                indexname = "taurusValueIndex" + str(nb_ref + 2)
                self.taurusValueIndex[nb_ref].setObjectName(indexname)
                self.taurusValueIndex[nb_ref].setValue(int(ref[0]))

                self.taurusValueH.append(TaurusValueLineEdit(w))
                self.taurusValueH[nb_ref].setGeometry(
                    QtCore.QRect(xh, 100 + 30 * (nb_ref), 81, 27))
                self.taurusValueH[nb_ref].setReadOnly(True)
                hname = "taurusValueH" + str(nb_ref + 2)
                self.taurusValueH[nb_ref].setObjectName(hname)
                self.taurusValueH[nb_ref].setValue("%10.4f" % ref[1])

                self.taurusValueK.append(TaurusValueLineEdit(w))
                self.taurusValueK[nb_ref].setGeometry(
                    QtCore.QRect(xk, 100 + 30 * (nb_ref), 81, 27))
                self.taurusValueK[nb_ref].setReadOnly(True)
                kname = "taurusValueK" + str(nb_ref + 2)
                self.taurusValueK[nb_ref].setObjectName(kname)
                self.taurusValueK[nb_ref].setValue("%10.4f" % ref[2])

                self.taurusValueL.append(TaurusValueLineEdit(w))
                self.taurusValueL[nb_ref].setGeometry(
                    QtCore.QRect(xl, 100 + 30 * (nb_ref), 81, 27))
                self.taurusValueL[nb_ref].setReadOnly(True)
                lname = "taurusValueL" + str(nb_ref + 2)
                self.taurusValueL[nb_ref].setObjectName(lname)
                self.taurusValueL[nb_ref].setValue("%10.4f" % ref[3])

                self.taurusValueRelevance.append(TaurusValueLineEdit(w))
                self.taurusValueRelevance[nb_ref].setGeometry(
                    QtCore.QRect(xrelevance, 100 + 30 * (nb_ref), 41, 27))
                self.taurusValueRelevance[nb_ref].setReadOnly(True)
                relevancename = "taurusValueRelevance" + str(nb_ref + 2)
                self.taurusValueRelevance[nb_ref].setObjectName(relevancename)
                self.taurusValueRelevance[nb_ref].setValue(int(ref[4]))

                self.taurusValueAffinement.append(TaurusValueLineEdit(w))
                self.taurusValueAffinement[nb_ref].setGeometry(
                    QtCore.QRect(xaffinement, 100 + 30 * (nb_ref), 41, 27))
                self.taurusValueAffinement[nb_ref].setReadOnly(True)
                affinementname = "taurusValueAffinement" + str(nb_ref + 2)
                self.taurusValueAffinement[
                    nb_ref].setObjectName(affinementname)
                self.taurusValueAffinement[nb_ref].setValue(int(ref[5]))

                self.taurusValueAngle1.append(TaurusValueLineEdit(w))
                self.taurusValueAngle1[nb_ref].setGeometry(
                    QtCore.QRect(xangle1, 100 + 30 * (nb_ref), 81, 27))
                self.taurusValueAngle1[nb_ref].setReadOnly(True)
                angle1name = "taurusValueAngle1" + str(nb_ref + 2)
                self.taurusValueAngle1[nb_ref].setObjectName(angle1name)
                self.taurusValueAngle1[nb_ref].setValue("%10.4f" % ref[6])

                self.taurusValueAngle2.append(TaurusValueLineEdit(w))
                self.taurusValueAngle2[nb_ref].setGeometry(
                    QtCore.QRect(xangle2, 100 + 30 * (nb_ref), 81, 27))
                self.taurusValueAngle2[nb_ref].setReadOnly(True)
                angle2name = "taurusValueAngle2" + str(nb_ref + 2)
                self.taurusValueAngle2[nb_ref].setObjectName(angle2name)
                self.taurusValueAngle2[nb_ref].setValue("%10.4f" % ref[7])

                self.taurusValueAngle3.append(TaurusValueLineEdit(w))
                self.taurusValueAngle3[nb_ref].setGeometry(
                    QtCore.QRect(xangle3, 100 + 30 * (nb_ref), 81, 27))
                self.taurusValueAngle3[nb_ref].setReadOnly(True)
                angle3name = "taurusValueAngle3" + str(nb_ref + 2)
                self.taurusValueAngle3[nb_ref].setObjectName(angle3name)
                self.taurusValueAngle3[nb_ref].setValue("%10.4f" % ref[8])

                self.taurusValueAngle4.append(TaurusValueLineEdit(w))
                self.taurusValueAngle4[nb_ref].setGeometry(
                    QtCore.QRect(xangle4, 100 + 30 * (nb_ref), 81, 27))
                self.taurusValueAngle4[nb_ref].setReadOnly(True)
                angle4name = "taurusValueAngle4" + str(nb_ref + 2)
                self.taurusValueAngle4[nb_ref].setObjectName(angle4name)
                self.taurusValueAngle4[nb_ref].setValue("%10.4f" % ref[9])

                if len(ref) == 12:
                    self.taurusValueAngle5.append(TaurusValueLineEdit(w))
                    self.taurusValueAngle5[nb_ref].setGeometry(
                        QtCore.QRect(xangle5, 100 + 30 * (nb_ref), 81, 27))
                    self.taurusValueAngle5[nb_ref].setReadOnly(True)
                    angle5name = "taurusValueAngle5" + str(nb_ref + 2)
                    self.taurusValueAngle5[nb_ref].setObjectName(angle5name)
                    self.taurusValueAngle5[nb_ref].setValue("%10.4f" % ref[10])

                    self.taurusValueAngle6.append(TaurusValueLineEdit(w))
                    self.taurusValueAngle6[nb_ref].setGeometry(
                        QtCore.QRect(xangle6, 100 + 30 * (nb_ref), 81, 27))
                    self.taurusValueAngle6[nb_ref].setReadOnly(True)
                    angle6name = "taurusValueAngle6" + str(nb_ref + 2)
                    self.taurusValueAngle6[nb_ref].setObjectName(angle6name)
                    self.taurusValueAngle6[nb_ref].setValue("%10.4f" % ref[11])

                nb_ref = nb_ref + 1

            w.resize(930, 100 + nb_ref * 50)

        else:
            self.rl_label_nor = QtGui.QLabel(w)
            self.rl_label_nor.setGeometry(
                QtCore.QRect(xangle1 - 50, 110, 300, 20))
            font = QtGui.QFont()
            font.setPointSize(12)
            font.setWeight(75)
            font.setBold(True)
            self.rl_label_nor.setFont(font)
            self.rl_label_nor.setObjectName("rl_label_nor")
            self.rl_label_nor.setText(QtGui.QApplication.translate(
                "Form", "NO REFLECTIONS", None))

        w.show()
        w.show()

    def edit_reflections_window(self):

        w = ReflectionsEditor()
        w.setModel(self.model)

        w.show()

    def add_select_crystal(self):
        new_crystal = str(self._ui.NewCrystalLineEdit.text())
        self.device.write_attribute("AddCrystal", new_crystal)
        self.crystalscombobox.loadItems(self.device.crystallist)
        self.device.write_attribute("Crystal", new_crystal)

    def affine(self):
        self.device.write_attribute("affinecrystal", 1)


def main():
    parser = taurus.core.util.argparse.get_taurus_parser()
    parser.usage = "%prog  <model>"
    parser.set_description(
        "a taurus application for setting diffractometer parameters: "
        "ubmatrix, lattice, reflections, ...")

    app = taurus.qt.qtgui.application.TaurusApplication(
        cmd_line_parser=parser, app_version=sardana.Release.version)
    app.setApplicationName("ubmatrix")
    args = app.get_command_line_args()
    if len(args) < 1:
        msg = "model not set (requires diffractometer controller)"
        parser.error(msg)

    w = UBMatrixBase()
    w.model = args[0]
    w.setModel(w.model)
    w.show()

    sys.exit(app.exec_())

if __name__ == "__main__":
    main()
